#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   job.py
@Author  :   Wei Loon Cheng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Datature Job API
"""
# pylint: disable=E1102

import json
import logging
import os
import time
from datetime import datetime, timedelta
from typing import Dict, Generator, Optional, Union

import numpy as np
import requests
from alive_progress import alive_bar

from datature.nexus import config, error, models
from datature.nexus.api.batch import models as batch_models
from datature.nexus.api.batch.types import (
    JobMetadata,
    JobOptions,
    JobStatusOverview,
    LogsFilter,
)
from datature.nexus.api.operation import Operation
from datature.nexus.api.types import DEFAULT_INSTANCES, Pagination
from datature.nexus.client_context import ClientContext, RestContext

logger = logging.getLogger("datature-nexus")


class Job(RestContext):
    """Datature Job API Resource."""

    def __init__(self, client_context: ClientContext):
        """Initialize the API Resource."""
        super().__init__(client_context)
        self.operation = Operation(client_context)

    def cancel(self, job_id: str) -> batch_models.Job:
        """Cancel a running or pending job.

        :param job_id: The job ID.
        :return: A msgspec struct containing the job metadata with the following structure:

            .. code-block:: python

                Job(
                    id='batchjob_4e5566d0-6538-441a-9fa5-f3495516646a',
                    object='batch_job',
                    name='cli-test',
                    project_id='proj_ca5fe71e7592bbcf7705ea36e4f29ed4',
                    spec=JobSpec(
                        start_at_time=1723734419985,
                        stop_at_time=1723993619985,
                        dataset_id='dataset_652ef566-af5c-4d52-b1e4-ec8ec6dc4b8e',
                        result_delivery=ResultDelivery(
                            destinations=[Webhook(
                                webhook_id='webhook_ed981529-0160-4704-8794-e7702a8de470'
                            )]
                        ),
                        deployment=ExistingDeployment(
                            deployment_id='deploy_3286bd71-9a8b-42bb-8ba6-c07549b81367'
                        ),
                    ),
                    status=JobStatus(
                        overview='Reached',
                        message='Finished',
                        update_date=1723734509417,
                        items=JobItemStatus(
                            gathered=50,
                            failed_gather=0,
                            preprocessed=50,
                            failed_process=0,
                            predicted=50,
                            failed_predict=0,
                            delivered=50,
                            failed_deliver=0
                        ),
                        reason=None
                    ),
                    create_date=1723734419985,
                    update_date=1723734509434
                )

        :example:
            .. code-block:: python

                from datature.nexus import Client
                project = Client("5aa41e8ba........").get_project("proj_b705a........")

                project.batch.jobs.cancel("batchjob_4e5566d0-6538-441a-9fa5-f3495516646a")
        """
        assert isinstance(job_id, str)
        return self.requester.PATCH(
            f"/projects/{self.project_id}/batch/jobs/{job_id}",
            request_body={"spec": {"stopAtTime": int(time.time() * 1000)}},
            response_type=batch_models.Job,
        )

    def create(self, name: str, job_options: dict) -> batch_models.Job:
        """Create a new batch job.

        :param name: The name of the job.
        :param job_options: The job options.
        :returns: A msgspec struct containing the job metadata with the following structure:

            .. code-block:: python

                Job(
                    id='batchjob_4e5566d0-6538-441a-9fa5-f3495516646a',
                    object='batch_job',
                    name='cli-test',
                    project_id='proj_ca5fe71e7592bbcf7705ea36e4f29ed4',
                    spec=JobSpec(
                        start_at_time=1723734419985,
                        stop_at_time=1723993619985,
                        dataset_id='dataset_652ef566-af5c-4d52-b1e4-ec8ec6dc4b8e',
                        result_delivery=ResultDelivery(
                            destinations=[Webhook(
                                webhook_id='webhook_ed981529-0160-4704-8794-e7702a8de470'
                            )]
                        ),
                        deployment=ExistingDeployment(
                            deployment_id='deploy_3286bd71-9a8b-42bb-8ba6-c07549b81367'
                        ),
                    ),
                    status=JobStatus(
                        overview='Reached',
                        message='Finished',
                        update_date=1723734509417,
                        items=JobItemStatus(
                            gathered=50,
                            failed_gather=0,
                            preprocessed=50,
                            failed_process=0,
                            predicted=50,
                            failed_predict=0,
                            delivered=50,
                            failed_deliver=0
                        ),
                        reason=None
                    ),
                    create_date=1723734419985,
                    update_date=1723734509434
                )

        :example:
            .. code-block:: python

                from datature.nexus import Client
                project = Client("5aa41e8ba........").get_project("proj_b705a........")

                project.batch.jobs.create(
                    "my-batch-job-08-15",
                    {
                        "dataset_id": "dataset_652ef566-af5c-4d52-b1e4-ec8ec6dc4b8e",
                        "webhook_id"="webhook_ed981529-0160-4704-8794-e7702a8de470",
                        "deployment_id": "deploy_3286bd71-9a8b-42bb-8ba6-c07549b81367",
                        "start_at_time": 1723734419985,
                        "stop_at_time": 1723993619985,
                    }
                )
        """
        assert isinstance(job_options, dict)

        if job_options.get("accumulate_results"):
            return self._create_job_with_accumulated_results(job_options)

        deployment_metadata = job_options.get("deployment_metadata")
        if deployment_metadata:
            if not deployment_metadata.get(
                "instance_id"
            ) and not deployment_metadata.get("resources"):
                logger.warning(
                    "WARNING: The `instance_id` field is not set. A random default instance will be selected. "
                    "To view the list of default instances, refer to our documentation: "
                    "https://developers.datature.io/docs/deployment-configuration#instance-types"
                )

                region = job_options.get("deployment_metadata").get("region") or "*"
                available_instances = self.requester.GET(
                    f"/projects/{self.project_id}/artifacts/{job_options.get('artifact_id')}"
                    f"/availableInstances/regions/{region}",
                    response_type=models.AvailableInstances,
                ).instances

                default_instances = [
                    instance.id
                    for instance in available_instances
                    if instance.id in DEFAULT_INSTANCES
                ]

                job_options["deployment_metadata"]["instance_id"] = np.random.choice(
                    default_instances
                )

        job_metadata = JobMetadata(
            name=name, spec=JobOptions(**job_options).parse_to_spec()
        )

        return self.requester.POST(
            f"/projects/{self.project_id}/batch/jobs",
            request_body=job_metadata.to_json(),
            response_type=batch_models.Job,
        )

    def _create_job_with_accumulated_results(
        self, job_options: dict
    ) -> batch_models.Job:
        """Create a new batch job with accumulated results.

        :param job_options: The job options.
        :returns: A dictionary containing the batch job ID.
        """
        batch_job_accumulator_url = os.getenv(
            "DATATURE_API_BATCH_JOB_ACCUMULATOR_URL",
            "https://monopod-prod-384071953537.asia-southeast1.run.app",
        )

        data = {
            "dt_project_key": self.project_id.split("_")[-1],
            "dt_secret_key": self._context.secret_key,
            "ndjson_url": job_options.get("ndjson_url"),
            "artifact_ids": job_options.get("artifact_ids"),
        }

        try:
            job_response = requests.request(
                "POST",
                f"{batch_job_accumulator_url}/create-batchjob",
                headers={"Content-Type": "application/json"},
                data=json.dumps(data),
                timeout=config.REQUEST_TIME_OUT_SECONDS,
            )
        except requests.exceptions.RequestException as exc:
            logger.error("Failed to create job with accumulated results: %s", exc)
            raise error.BadRequestError(
                "Failed to create job with accumulated results."
            )

        return job_response.json()

    def delete(self, job_id: str) -> models.DeleteResponse:
        """Delete a batch job.

        :param job_id: The job ID.
        :returns: A msgspec struct containing the delete response with the following structure:

            .. code-block:: python

                DeleteResponse(
                    id='batchjob_4e5566d0-6538-441a-9fa5-f3495516646a',
                    deleted=True
                )

        :example:
            .. code-block:: python

                from datature.nexus import Client
                project = Client("5aa41e8ba........").get_project("proj_b705a........")

                project.batch.jobs.delete("batchjob_4e5566d0-6538-441a-9fa5-f3495516646a")
        """
        assert isinstance(job_id, str)
        return self.requester.DELETE(
            f"/projects/{self.project_id}/batch/jobs/{job_id}",
            response_type=models.DeleteResponse,
        )

    def get(self, job_id: str) -> batch_models.Job:
        """Get a batch job.

        :param job_id: The job ID.
        :returns: A msgspec struct containing the job metadata with the following structure:

            .. code-block:: python

                Job(
                    id='batchjob_4e5566d0-6538-441a-9fa5-f3495516646a',
                    object='batch_job',
                    name='my-batch-job-08-15',
                    project_id='proj_ca5fe71e7592bbcf7705ea36e4f29ed4',
                    spec=JobSpec(
                        start_at_time=1723734419985,
                        stop_at_time=1723993619985,
                        dataset_id='dataset_652ef566-af5c-4d52-b1e4-ec8ec6dc4b8e',
                        result_delivery=ResultDelivery(
                            destinations=[Webhook(
                                webhook_id='webhook_ed981529-0160-4704-8794-e7702a8de470'
                            )]
                        ),
                        deployment=ExistingDeployment(
                            deployment_id='deploy_3286bd71-9a8b-42bb-8ba6-c07549b81367'
                        ),
                    ),
                    status=JobStatus(
                        overview='Reached',
                        message='Finished',
                        update_date=1723734509417,
                        items=JobItemStatus(
                            gathered=50,
                            failed_gather=0,
                            preprocessed=50,
                            failed_process=0,
                            predicted=50,
                            failed_predict=0,
                            delivered=50,
                            failed_deliver=0
                        ),
                        reason=None
                    ),
                    create_date=1723734419985,
                    update_date=1723734509434
                )

        :example:
            .. code-block:: python

                from datature.nexus import Client
                project = Client("5aa41e8ba........").get_project("proj_b705a........")

                project.batch.jobs.get("batchjob_4e5566d0-6538-441a-9fa5-f3495516646a")
        """
        assert isinstance(job_id, str)
        return self.requester.GET(
            f"/projects/{self.project_id}/batch/jobs/{job_id}",
            response_type=batch_models.Job,
        )

    def list(
        self,
        pagination: Union[Pagination, dict, None] = None,
    ) -> models.PaginationResponse[batch_models.Job]:
        """List all batch jobs.

        :param pagination: The pagination options.
        :returns: A msgspec struct containing the pagination response with the following structure:

            .. code-block:: python

                PaginationResponse(
                    next_page=None,
                    prev_page=None,
                    data=[
                        Job(
                            id='batchjob_4e5566d0-6538-441a-9fa5-f3495516646a',
                            object='batch_job',
                            name='my-batch-job-08-15',
                            project_id='proj_ca5fe71e7592bbcf7705ea36e4f29ed4',
                            spec=JobSpec(
                                start_at_time=1723734419985,
                                stop_at_time=1723993619985,
                                dataset_id='dataset_652ef566-af5c-4d52-b1e4-ec8ec6dc4b8e',
                                result_delivery=ResultDelivery(
                                    destinations=[Webhook(
                                        webhook_id='webhook_ed981529-0160-4704-8794-e7702a8de470'
                                    )]
                                ),
                                deployment=ExistingDeployment(
                                    deployment_id='deploy_3286bd71-9a8b-42bb-8ba6-c07549b81367'
                                ),
                            ),
                            status=JobStatus(
                                overview='Reached',
                                message='Finished',
                                update_date=1723734509417,
                                items=JobItemStatus(
                                    gathered=50,
                                    failed_gather=0,
                                    preprocessed=50,
                                    failed_process=0,
                                    predicted=50,
                                    failed_predict=0,
                                    delivered=50,
                                    failed_deliver=0
                                ),
                                reason=None
                            ),
                            create_date=1723734419985,
                            update_date=1723734509434
                        )
                    ],
                )

        :example:
            .. code-block:: python

                from datature.nexus import Client
                project = Client("5aa41e8ba........").get_project("proj_b705a........")

                project.batch.jobs.list()
        """
        assert isinstance(pagination, (Pagination, dict, type(None)))

        if isinstance(pagination, dict):
            pagination = Pagination(**pagination)
        if pagination is None:
            pagination = Pagination()

        return self.requester.GET(
            f"/projects/{self.project_id}/batch/jobs",
            query={**pagination.to_json()},
            response_type=models.PaginationResponse[batch_models.Job],
        )

    def get_logs(
        self, job_id: str, logs_filter: Union[LogsFilter, dict, None] = None
    ) -> batch_models.JobLogs:
        """Get the logs for a batch job. This query is limited to a maximum of 1000 entries.
        To get more logs, use the `get_all_log_entries` method.

        :param job_id: The job ID.
        :param logs_filter: The logs filter.
        :returns: A msgspec struct containing the job logs with the following structure:

            .. code-block:: python

                JobLogs(
                    id='batchjoblog_3d561cfa-9a53-4eb3-81c1-4193ae27f4fc',
                    object='batch_job_log',
                    project_id='proj_ca5fe71e7592bbcf7705ea36e4f29ed4',
                    entries=[
                        JobLogEntry(
                            kind='ItemPredicted',
                            create_time=1723080300011,
                            id='1e0ad7896a4aa2dd47eda727',
                            level='Info',
                            item='https://my-endpoint.url/blood_cell/36.jpg',
                            status='Delivered',
                            time_ms=9819.226,
                            reason=''
                        )
                    ]
                )

        :example:
            .. code-block:: python

                from datature.nexus import Client
                project = Client("5aa41e8ba........").get_project("proj_b705a........")

                project.batch.jobs.get_logs("batchjob_4e5566d0-6538-441a-9fa5-f3495516646a")
        """
        assert isinstance(logs_filter, (LogsFilter, dict, type(None)))

        if isinstance(logs_filter, dict):
            logs_filter = LogsFilter(**logs_filter)
        if logs_filter is None:
            logs_filter = LogsFilter()

        return self.requester.GET(
            f"/projects/{self.project_id}/batch/jobs/{job_id}/logs",
            query={**logs_filter.to_json()},
            response_type=batch_models.JobLogs,
        )

    def get_all_log_entries(
        self, job_id: str, logs_filter: Union[LogsFilter, dict, None] = None
    ) -> Generator[batch_models.JobLogEntry, None, None]:
        """Get all the logs for a batch job in batches of 1000 entries and return a generator.

        :param job_id: The job ID.
        :param logs_filter: The logs filter.
        :returns: A generator of msgspec struct containing the
                  job log entries with the following structure:

            .. code-block:: python

                JobLogEntry(
                    kind='ItemPredicted',
                    create_time=1723080300011,
                    id='1e0ad7896a4aa2dd47eda727',
                    level='Info',
                    item='https://my-endpoint.url/blood_cell/36.jpg',
                    status='Delivered',
                    time_ms=9819.226,
                    reason=''
                )

        :example:
            .. code-block:: python

                from datature.nexus import Client
                project = Client("5aa41e8ba........").get_project("proj_b705a........")

                count = 0
                for entry in project.batch.jobs.get_all_log_entries(
                    "batchjob_4e5566d0-6538-441a-9fa5-f3495516646a"
                ):
                    count += 1
                print(f"Total log entries: {count}")
        """
        with alive_bar(
            title="Fetching Logs",
            length=15,
            unit=" entries",
            force_tty=True,
            enrich_print=False,
        ) as progress_bar:
            seen_ids = set()
            while entries := self.get_logs(job_id, logs_filter).entries:
                if not entries:
                    break

                for entry in entries:
                    if entry.id not in seen_ids:
                        progress_bar()
                        yield entry
                seen_ids.update(entry.id for entry in entries)

                if len(entries) < 1000:
                    break

                if isinstance(logs_filter, dict):
                    logs_filter = LogsFilter(**logs_filter)
                elif logs_filter is None:
                    logs_filter = LogsFilter()
                logs_filter.before_time = entries[-1].create_time

    def wait_until_done(
        self,
        job_id: str,
        interval: int = config.OPERATION_LOOPING_DELAY_SECONDS,
        timeout: int = config.OPERATION_LOOPING_TIMEOUT_SECONDS,
        raise_exception_if: Union[JobStatusOverview, str] = JobStatusOverview.FAILED,
    ) -> batch_models.Job:
        """Wait until a batch job is done.

        :param job_id: The job ID.
        :param interval: The interval to check the job status.
        :param timeout: The timeout to wait for the job to be completed.
        :param raise_exception_if: The status to raise an exception if reached.
        :returns: A msgspec struct containing the job metadata with the following structure:

            .. code-block:: python

                Job(
                    id='batchjob_4e5566d0-6538-441a-9fa5-f3495516646a',
                    object='batch_job',
                    name='my-batch-job-08-15',
                    project_id='proj_ca5fe71e7592bbcf7705ea36e4f29ed4',
                    spec=JobSpec(
                        start_at_time=1723734419985,
                        stop_at_time=1723993619985,
                        dataset_id='dataset_652ef566-af5c-4d52-b1e4-ec8ec6dc4b8e',
                        result_delivery=ResultDelivery(
                            destinations=[Webhook(
                                webhook_id='webhook_ed981529-0160-4704-8794-e7702a8de470'
                            )]
                        ),
                        deployment=ExistingDeployment(
                            deployment_id='deploy_3286bd71-9a8b-42bb-8ba6-c07549b81367'
                        ),
                    ),
                    status=JobStatus(
                        overview='Reached',
                        message='Finished',
                        update_date=1723734509417,
                        items=JobItemStatus(
                            gathered=50,
                            failed_gather=0,
                            preprocessed=50,
                            failed_process=0,
                            predicted=50,
                            failed_predict=0,
                            delivered=50,
                            failed_deliver=0
                        ),
                        reason=None
                    ),
                    create_date=1723734419985,
                    update_date=1723734509434
                )

        :example:
            .. code-block:: python

                from datature.nexus import Client
                project = Client("5aa41e8ba........").get_project("proj_b705a........")

                project.batch.jobs.wait_until_done(
                    "batchjob_4e5566d0-6538-441a-9fa5-f3495516646a"
                )
        """
        assert isinstance(interval, int) and interval > 0
        assert isinstance(timeout, int) and timeout > 0
        assert isinstance(raise_exception_if, (JobStatusOverview, str))

        if isinstance(raise_exception_if, str):
            raise_exception_if = JobStatusOverview(raise_exception_if)

        elapsed_time = datetime.now() + timedelta(seconds=timeout)
        response = self.get(job_id)
        delivered_and_failed_count = (
            response.status.items.delivered + response.status.items.failed_deliver
        )

        dataset_id = response.spec.dataset_id
        dataset_count = self.requester.GET(
            f"/projects/{self.project_id}/batch/datasets/{dataset_id}",
            response_type=batch_models.Dataset,
        ).status.item_count

        with alive_bar(
            dataset_count,
            title="Processing Batch Job",
            length=20,
            force_tty=True,
            enrich_print=False,
        ) as progress_bar:
            cumulative_item_count = 0
            while delivered_and_failed_count != dataset_count:
                if (
                    response.status.overview == raise_exception_if.value
                    and delivered_and_failed_count != dataset_count
                ):
                    raise error.BadRequestError(
                        f"Error processing Job {job_id}, "
                        f"run `project.batch.jobs.get({job_id})` in Python or "
                        f"`datature batch jobs get {job_id}` in your terminal to get more detailed "
                        "status messages, or contact support at support@datature.io",
                    )

                if elapsed_time < datetime.now():
                    logger.warning(
                        "Operation timeout: please use `project.batch.jobs.get(%s)` in Python or "
                        "`datature batch jobs get %s` in your terminal to get the status instead.",
                        job_id,
                        job_id,
                    )
                    return response

                new_progress_count = delivered_and_failed_count - cumulative_item_count
                if new_progress_count > 0:
                    progress_bar(new_progress_count)
                    cumulative_item_count = delivered_and_failed_count
                time.sleep(interval)

                response = self.get(job_id)
                delivered_and_failed_count = (
                    response.status.items.delivered
                    + response.status.items.failed_deliver
                )
            if progress_bar.current != dataset_count:
                progress_bar(dataset_count - progress_bar.current)

        # Final sleep to allow a small buffer for the status to be updated
        time.sleep(interval)
        response = self.get(job_id)
        if response.status.items.delivered != dataset_count:
            logger.warning(
                "Job has completed but not all items have been successful. "
                "Run `project.batch.jobs.get_logs(%s)` in Python or "
                "`datature batch jobs logs %s` in your terminal to check the logs "
                "for more details, or contact support at support@datature.io",
                job_id,
                job_id,
            )
        return response

    def get_accumulated_results(self, job_id: str) -> Optional[Dict[str, str]]:
        """Get the accumulated results for a batch job.

        :param job_id: The job ID.
        :returns: The accumulated results.

        :example:
            .. code-block:: python

                from datature.nexus import Client
                project = Client("5aa41e8ba........").get_project("proj_b705a........")

                project.batch.jobs.get_accumulated_results("batchjob_4e5566d0-6538-441a-9fa5-f3495516646a")
        """
        batch_job_accumulator_url = os.getenv(
            "DATATURE_API_BATCH_JOB_ACCUMULATOR_URL",
            "https://monopod-prod-384071953537.asia-southeast1.run.app",
        )
        response = requests.request(
            "GET",
            f"{batch_job_accumulator_url}/retrieve-result",
            headers={"Content-Type": "application/json"},
            params={"batch_job_id": job_id},
            timeout=config.REQUEST_TIME_OUT_SECONDS,
        )
        return response.json()



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   error.py
@Author  :   Raighne.Weng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Error Class module
"""


class Error(Exception):
    """Error Class."""

    def __init__(
        self,
        message,
        detail=None,
    ):
        super()
        self.message = message
        self.detail = detail


class ErrorWithCode(Error):
    """HTTP Error Class."""

    def __repr__(self):
        return (
            f"{self.__class__.__name__}"
            f"(message={self.message}, "
            f"detail={self.detail})"
        )


class BadRequestError(ErrorWithCode):
    """BadRequestError http_code: 400"""


class UnauthorizedError(ErrorWithCode):
    """UnauthorizedError http_code: 401"""


class ForbiddenError(ErrorWithCode):
    """ForbiddenError http_code: 403"""


class NotFoundError(ErrorWithCode):
    """NotFoundError http_code: 404"""


class TooManyRequestsError(ErrorWithCode):
    """TooManyRequestsError http_code: 429"""


class InternalServerError(ErrorWithCode):
    """InternalServerError http_code: 500"""



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   artifact_fixture.py
@Author  :   Raighne.Weng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Datature Artifact Test Data
"""

# pylint: disable=C0301
artifacts_includes_exports_response = [
    {
        "id": "artifact_652e7022a70966c12f2203c1",
        "object": "artifact",
        "runId": "run_5cadcd5b-1590-4f2e-83b8-cb908c57087c",
        "projectId": "proj_ff801f417e7199e46a5c1eaa22052d56",
        "isTraining": False,
        "step": 5000,
        "flowTitle": "Yolov8 Workflow",
        "artifactName": "ckpt-23-datature-yolov8n-cls",
        "modelName": "fasterrcnn-inceptionv2-1024x1024",
        "createDate": 1697542178156,
        "metric": {"totalLoss": 0.0032949},
        "isDeployed": False,
        "exportOptions": [
            {
                "format": "ONNX",
                "optimizations": {"quantization": ["float32", "float16"]},
                "defaultOptimizations": {"quantization": "float32"},
            },
            {
                "format": "TensorFlow",
                "optimizations": {"quantization": ["float32"]},
                "defaultOptimizations": {"quantization": "float32"},
            },
            {
                "format": "PyTorch",
                "optimizations": {"quantization": ["float32"]},
                "defaultOptimizations": {"quantization": "float32"},
            },
            {
                "format": "TFLite",
                "optimizations": {"quantization": ["float32"]},
                "defaultOptimizations": {"quantization": "float32"},
            },
        ],
    },
    {
        "id": "artifact_652e7022a70966c12f2203c0",
        "object": "artifact",
        "runId": "run_5cadcd5b-1590-4f2e-83b8-cb908c57087b",
        "projectId": "proj_ff801f417e7199e46a5c1eaa22052d56",
        "isTraining": False,
        "step": 5000,
        "flowTitle": "Yolov8 Workflow",
        "artifactName": "ckpt-23-datature-yolov8n-cls",
        "modelName": "fasterrcnn-inceptionv2-1024x1024",
        "createDate": 1697542178156,
        "metric": {"totalLoss": 0.0032949},
        "isDeployed": False,
        "exportOptions": [
            {
                "format": "ONNX",
                "optimizations": {"quantization": ["float32", "float16"]},
                "defaultOptimizations": {"quantization": "float32"},
            },
            {
                "format": "TensorFlow",
                "optimizations": {"quantization": ["float32"]},
                "defaultOptimizations": {"quantization": "float32"},
            },
            {
                "format": "PyTorch",
                "optimizations": {"quantization": ["float32"]},
                "defaultOptimizations": {"quantization": "float32"},
            },
            {
                "format": "TFLite",
                "optimizations": {"quantization": ["float32"]},
                "defaultOptimizations": {"quantization": "float32"},
            },
        ],
        "exports": [
            {
                "id": "model_x4qq70769621y5rrv007447q7y5xw279",
                "object": "model",
                "artifactId": "artifact_652e7022a70966c12f2203c0",
                "status": "Finished",
                "format": "Onnx",
                "quantization": "float32",
                "createDate": 1699584021867,
                "download": {
                    "method": "GET",
                    "expiryDate": 1706263764534,
                    "url": "https://storage.googleapis.com/exports.staging.datature.io/ff801f417e7199e46a5c1eaa22052d56/models/5cadcd5b-1590-4f2e-83b8-cb908c57087b-ckpt-23-datature-yolov8n-cls-onnx.zip?X-Goog-Algorithm=GOOG4-RSA-SHA256&X-Goog-Credential=artifact-services%40datature-nexus-staging.iam.gserviceaccount.com%2F20240125%2Fauto%2Fstorage%2Fgoog4_request&X-Goog-Date=20240125T100924Z&X-Goog-Expires=86400&X-Goog-SignedHeaders=host&X-Goog-Signature=17b63f5804b6dcf20693e956704151d0c0720df7c182833ff96c67219b2f8966274008ba872192c160fc0b00ac11f39789bf01152d03e7a81185903d7f4c027d5675795dfcae36ebb93709fca96bc1872993c1f251ac7cadbeb14fcd5e112c4df8af053462072db403b20d551019c5248ac94d0c7cda652db40033d4adeefa882e5f6b52de1153426a4b08d8e54560b7a02b1769171ea2a02fd9f0bf91cd4f877a25e538316222f0eb534a28e038247786a07c552c515229404d2aa02264807bbe3b6163485770deeb90934b22be4c7cdfcfd8ba436f429de45bfc7f9131e576fb75181ab1bcfec8054753a0ba2633a2a696ce6aa6ea1a2812fcbbd26b0eec31",
                },
            }
        ],
    },
]



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   test_workflow.py
@Author  :   Raighne.Weng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Datature Workflow API Test Cases
"""

import unittest
from unittest.mock import ANY, MagicMock

from datature.nexus.api.workflow import Workflow
from datature.nexus.client_context import ClientContext

# pylint: disable=W0212


class TestWorkflow(unittest.TestCase):
    """Datature Workflow API Resource Test Cases."""

    def __init__(self, *args, **kwargs):
        """init global variables."""
        unittest.TestCase.__init__(self, *args, **kwargs)

        context = ClientContext("secret_key", endpoint="http://localhost:8080")
        context.project_id = "project_id"
        context.requester._request = MagicMock()

        self.workflow = Workflow(context)

    def test_list(self):
        """Test list all workflows."""
        self.workflow.list()

        self.workflow._context.requester._request.assert_called_once_with(
            "GET",
            "/projects/project_id/workflows",
            ANY,
            query=None,
            request_params=None,
        )

    def test_retrieve(self):
        """Test retrieve a workflow."""
        self.workflow.get("flow_id")

        self.workflow._context.requester._request.assert_called_once_with(
            "GET",
            "/projects/project_id/workflows/flow_id",
            ANY,
            query=None,
            request_params=None,
        )

    def test_modify(self):
        """Test update a workflow."""
        self.workflow.update("flow_id", {"title": "New Workflow Title"})

        self.workflow._context.requester._request.assert_called_once_with(
            "PATCH",
            "/projects/project_id/workflows/flow_id",
            ANY,
            request_body={"title": "New Workflow Title"},
            query=None,
            request_params=None,
        )

    def test_delete(self):
        """Test delete a workflow."""
        self.workflow.delete("flow_id")

        self.workflow._context.requester._request.assert_called_once_with(
            "DELETE", "/projects/project_id/workflows/flow_id", ANY
        )



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   batch.py
@Author  :   Wei Loon Cheng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Datature Batch API
"""

from collections import ChainMap
from inspect import isclass

from datature.nexus.api.batch.dataset import Dataset
from datature.nexus.api.batch.job import Job
from datature.nexus.api.batch.webhook import Webhook
from datature.nexus.client_context import ClientContext, RestContext


class Batch(RestContext):
    """Datature Batch API Resource."""

    datasets: Dataset
    jobs: Job
    webhooks: Webhook

    def __init__(self, client_context: ClientContext):
        """Initialize the Batch API Resource."""
        super().__init__(client_context)

        # Init sub resources
        self._auto_init_resources()

    def _get_all_annotations(self):
        """Get annotations for this class and all superclasses."""
        # This expression correctly handles cases where multiple superclasses
        # provide different annotations for the same attribute name.
        # See https://stackoverflow.com/a/72037059.
        return ChainMap(
            *(
                c.__annotations__
                for c in self.__class__.__mro__
                if "__annotations__" in c.__dict__
            )
        )

    def _auto_init_resources(self):
        for name, type_annot in self._get_all_annotations().items():
            # Don't overwrite existing fields
            if hasattr(self, name):
                continue

            # Only initialize if the type is a class and a subclass of RestContext
            if not isclass(type_annot) or not issubclass(type_annot, RestContext):
                continue

            setattr(self, name, type_annot(self._context))



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   upload_session.py
@Author  :   Raighne.Weng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Datature Asset Upload Session
"""
# pylint: disable=R1732, R0902, E0203, W0201

import concurrent.futures
import logging
import struct
import tempfile
import threading
from contextlib import ContextDecorator
from os import path
from pathlib import Path
from typing import List, Optional, Union

import crc32c
import cv2
from filetype import filetype

from datature.nexus import config, error
from datature.nexus.api.operation import Operation
from datature.nexus.api.types import OperationStatusOverview
from datature.nexus.client_context import ClientContext, RestContext
from datature.nexus.medical import get_processor
from datature.nexus.models import UploadSession as UploadSessionModel
from datature.nexus.utils import file_signature, utils

logger = logging.getLogger("datature-nexus")


class UploadSession(RestContext, ContextDecorator):
    """Datature Asset Upload Session Class.

    :param client_context: An instance of ClientContext.
    :param groups: A list of group names to categorize the upload. Default is None.
    :param background:
        A flag indicating whether the upload should run in the background. Default is False.
    """

    def __init__(
        self,
        client_context: ClientContext,
        groups: Optional[List[str]] = None,
        background: bool = False,
    ):
        """Initialize the API Resource."""
        super().__init__(client_context)
        self._local = threading.local()

        self._operation = None  # Lazy initialization if needed
        self.assets = []
        self.file_name_map = {}
        self.operation_ids = []

        self.groups = groups if groups is not None else ["main"]
        self.background = background

    @property
    def operation(self):
        """Initialize operation."""
        if self._operation is None:
            self._operation = Operation(self._context)
        return self._operation

    def _init_http_session(self):
        """Initialize local session and retry policy."""

        self._local.http_session = utils.init_gcs_upload_session()

    def __enter__(self):
        return self

    def __exit__(self, _exc_type, exc_val, _exc_tb):
        """Exit function.
        The function will be called if an exception is raised inside the context manager
        """
        if exc_val is not None:
            # Error handling, patch import session status
            logger.warning("Upload session error existed: %s", exc_val)
            raise error.Error(exc_val)

        # check asset length
        if len(self.assets) == 0 and len(self.operation_ids) == 0:
            raise error.Error("Assets to upload is empty")

        # call API to get signed url
        if self.assets:
            response = self._upload_assets()
            self.operation_ids.append(response.op_id)

        if self.background:
            return {"op_ids": self.operation_ids}

        # Wait server finish generate thumbnail
        self.wait_until_done()

        return {"op_ids": self.operation_ids}

    def __len__(self):
        """Over write len function."""
        return len(self.file_name_map)

    def add_path(self, file_path: str, custom_metadata: dict = None, **kwargs):
        """
        Add asset to upload.

        :param file_path: The path of the file to upload.
        :param custom_metadata: A dictionary of custom metadata to attach to the asset.
        :param kwargs: Additional keyword arguments to pass to the API.
        """
        if not path.exists(file_path):
            raise error.Error("Cannot find the Asset file")

        if path.isdir(file_path):
            file_paths = utils.find_all_assets(file_path)
        else:
            file_paths = [file_path]

        # Convert DICOM and NII file to video asset
        processed_files = self._pre_process_medical_file(file_paths, kwargs)

        for processed_file in processed_files:
            self._generate_metadata(
                processed_file.get("filename"),
                processed_file.get("file_path"),
                custom_metadata,
            )
            # check current asset size
            self._check_current_asset_size()

    def add_bytes(
        self, file_bytes: bytes, filename: str, custom_metadata: dict = None, **kwargs
    ):
        """Attach file in bytes to upload session

        :param file_bytes: The bytes of the file to upload.
        :param filename: The name of the file to upload, should include the file extension.
        :param custom_metadata: A dictionary of custom metadata to attach to the asset.
        :param kwargs: Additional keyword arguments to pass to the API.
        """
        file_mime_type = file_signature.get_file_mime_by_signature(file_bytes)

        if file_mime_type is None:
            raise TypeError(f"UnSupported file: {filename}")

        # Create a temporary directory
        temp_dir = tempfile.mkdtemp()

        # Create a temporary file path
        temp_file_path = path.join(temp_dir, filename)

        # Write file bytes to the temporary file
        with open(temp_file_path, "wb") as temp_file:
            temp_file.write(file_bytes)

        # Convert DICOM and NII file to video asset
        processed_files = self._pre_process_medical_file([temp_file_path], kwargs)

        for processed_file in processed_files:
            self._generate_metadata(
                processed_file.get("filename"),
                processed_file.get("file_path"),
                custom_metadata,
            )
            # check current asset size
            self._check_current_asset_size()

    def _pre_process_medical_file(self, file_paths: str, options: dict = None):
        """Pre process the file, if medical file convert to MP4 first."""
        pre_processed_medical_file = []

        for file_path in file_paths:
            if file_path.lower().endswith((".dcm", ".nii")):
                processor = get_processor(file_path)

                process_data = {"file": file_path, "options": options}
                processor.valid(process_data)
                resp = processor.process(process_data)

                if resp:
                    # change the converted file_path and filename
                    pre_processed_medical_file = pre_processed_medical_file + [
                        {"filename": Path(file_path).stem, "file_path": file_path}
                        for file_path in resp
                    ]
            else:
                pre_processed_medical_file.append(
                    {"filename": path.basename(file_path), "file_path": file_path}
                )
        return pre_processed_medical_file

    def _generate_metadata(
        self, filename: str, file_path: str, custom_metadata: dict = None
    ):
        """process the file to asset metadata."""
        with open(file_path, "rb") as file:
            contents = file.read()

            # calculate file crc32
            file_hash = crc32c.CRC32CHash()
            file_hash.update(contents)

            # To fix the wrong crc32 caused by mac M1 clip
            crc32 = struct.unpack(">l", file_hash.digest())[0]

            size = path.getsize(file_path)

            mime_kind = filetype.guess(file_path)

            file.close()

            if self.file_name_map.get(filename) is not None:
                raise error.Error(
                    f"Cannot add multiple files with the same name, {filename}"
                )

            if filename and size and crc32 and mime_kind:
                if mime_kind.mime in ["image/jpeg", "image/png"]:
                    asset_metadata = {
                        "filename": filename,
                        "size": size,
                        "crc32c": crc32,
                        "mime": mime_kind.mime,
                        "customMetadata": custom_metadata or {},
                    }
                elif mime_kind.mime == "video/mp4":
                    cap = cv2.VideoCapture(file_path)
                    frames = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))

                    asset_metadata = {
                        "filename": filename,
                        "size": size,
                        "crc32c": crc32,
                        "mime": mime_kind.mime,
                        "frames": frames,
                        "encoder": {"profile": "h264Saver", "everyNthFrame": 1},
                        "customMetadata": custom_metadata or {},
                    }
                else:
                    raise error.Error("UnSupported asset file")

                self.assets.append(asset_metadata)
                self.file_name_map[filename] = {"path": file_path}

                logger.debug("Add asset: %s", asset_metadata)
            else:
                raise error.Error("UnSupported asset file")

    def _upload_file_thought_signed_url(self, asset_upload):
        filename = asset_upload.get("metadata").get("filename")
        file_path = self.file_name_map.get(filename)["path"]

        # upload asset to GCP
        self._local.http_session.request(
            asset_upload.upload.method,
            asset_upload.upload.url,
            headers=asset_upload.upload.headers,
            data=open(file_path, "rb"),
            timeout=config.REQUEST_TIME_OUT_SECONDS,
        )
        return filename

    def _upload_assets(self):
        """Use ThreadPoolExecutor to upload files to GCP."""
        upload_session_response = self.requester.POST(
            f"/projects/{self.project_id}/assetuploadsessions",
            request_body={"groups": self.groups, "assets": self.assets},
            response_type=UploadSessionModel,
        )

        # Use ThreadPoolExecutor to upload files in parallel
        with concurrent.futures.ThreadPoolExecutor(
            initializer=self._init_http_session
        ) as executor:
            # Submit tasks to the executor
            futures = [
                executor.submit(self._upload_file_thought_signed_url, asset_upload)
                for asset_upload in upload_session_response.assets
            ]

            # Optionally, you can handle the results of the uploads here
            for future in futures:
                filename = future.result()

                logger.debug("Finished Uploading: % s", filename)

            logger.debug("All files uploaded")

        return upload_session_response

    def wait_until_done(
        self,
        raise_exception_if: Union[
            OperationStatusOverview, str
        ] = OperationStatusOverview.ERRORED,
    ):
        """
        Wait for all operations to be done.
        This function only works when background is set to False.
        It functions the same as Operation.wait_until_done.

        :param raise_exception_if: The condition to raise error.
        :return: The operation status metadata if the operation has finished,
        """
        assert isinstance(raise_exception_if, (str, OperationStatusOverview))

        if isinstance(raise_exception_if, str):
            raise_exception_if = OperationStatusOverview(raise_exception_if)

        if len(self.operation_ids) == 0:
            logger.debug("All operations finished")
            return True

        # Use ThreadPoolExecutor to upload files in parallel
        with concurrent.futures.ThreadPoolExecutor() as executor:
            # Submit tasks to the executor
            futures = [
                executor.submit(
                    self.operation.wait_until_done,
                    op_id,
                    raise_exception_if=raise_exception_if,
                )
                for op_id in self.operation_ids
            ]

            # Optionally, you can handle the results of the uploads here
            for future in futures:
                res = future.result()

                logger.debug("Finished operation: % s", res)

            logger.debug("All operations finished")
        return True

    # check current asset size

    def _check_current_asset_size(self):
        if len(self.assets) >= config.ASSET_UPLOAD_SESSION_BATCH_SIZE:
            upload_operation = self._upload_assets()

            self.operation_ids.append(upload_operation.op_id)
            # clear current batch
            self.assets = []

    def get_operation_ids(self):
        """
        A list of operation IDs. Because some dependency limits,
        each operation allows a maximum of 5000 assets.
        So if the total number of assets goes up over 5000,
        it will return a list of operation IDs.

        If you want to control the operations manually,
        you can use this function to get the operation ids.
        And the call project.operation.wait_until_done or project.operation.get
        to wait for the operations to finish.

        :return: A list of operation ids.

        :example:
            .. code-block:: python

                ['op_1', 'op_2', 'op_3']

        """
        return self.operation_ids



#!/usr/env/bin python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   predictor.py
@Author  :   Wei Loon Cheng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   TensorRT predictor module.
"""

# pylint: disable=E1121,R0902,R0903,R0913,R0914

import glob
import os
from abc import ABC, abstractmethod
from typing import Dict, Tuple

import cv2
import numpy as np

try:
    import tritonclient.http as httpclient
except ImportError as import_exc:
    raise ModuleNotFoundError(
        "Triton Inference Server not installed! "
        "Please visit https://developers.datature.io/docs/tensorrt-export "
        "to learn how to set up your environment."
    ) from import_exc
from PIL import Image

from .logger import trt_logger
from .prediction_utils import (
    clamp,
    draw_bboxes,
    draw_class_labels,
    draw_instance_masks,
    draw_keypoints,
    draw_semantic_masks,
    draw_yolov8_instance_masks,
    get_model_config,
    load_label_map,
    load_skeleton,
    nms_boxes,
    postprocess_predictions,
    xywh2xyxy,
)

# Default Triton Inference Server URL
# if not set, it will default to localhost:8000
TRITON_DEFAULT_URL = "localhost:8000"


class BasePredictor(ABC):
    """Base class for all predictors"""

    def __init__(
        self,
        dtype: str,
        input_path: str,
        model_name: str,
        label_map_path: str,
        threshold: float = 0.7,
        save: bool = True,
        output_path: str = "",
        **kwargs,
    ) -> None:
        """Initialize the predictor.

        Args:
            dtype (str): The data type of the input image.
            input_path (str): The path to the input image.
            model_name (str): The name of the model.
            label_map_path (str): The path to the label map.
            threshold (float): The confidence threshold for object detection, defaults to 0.7.
            save (bool): Whether to save the output, defaults to True.
            output_path (str): The path to the output directory, defaults to "".
        """
        self._dtype = dtype
        self._input_path = input_path
        self._model_name = model_name
        self._label_map_path = label_map_path
        self._skeleton_file_path = kwargs.get("skeleton_file_path")
        self._threshold = threshold
        self._save = save
        self._output_path = output_path
        self._model_config = get_model_config(self._model_name)
        trt_logger.debug("Model config:")
        trt_logger.debug(self._model_config)

        self._triton_client = httpclient.InferenceServerClient(url=TRITON_DEFAULT_URL)

        ## Load label map
        self._category_index = load_label_map(self._label_map_path)

        ## Load color map
        self._color_map = {}
        for each_class in self._category_index:
            self._color_map[each_class] = np.array(
                [int(i) for i in np.random.choice(range(256), size=3)], dtype=np.uint8
            )
        self._color_map[0] = np.array([0, 0, 0], dtype=np.uint8)

    @abstractmethod
    def __call__(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def _preprocess(self, path: str) -> np.ndarray:
        raise NotImplementedError

    @abstractmethod
    def _postprocess(self, response_outputs, num_classes: int):
        raise NotImplementedError

    @abstractmethod
    def _run_inference_client(self, model_input: np.ndarray) -> Dict:
        raise NotImplementedError

    def _predict(self, image: np.ndarray) -> Tuple[np.ndarray]:
        """Send an image for prediction, then process it to be returned
        afterwards.

        Args:
            image (np.ndarray): The image to be predicted.

        Returns:
            Tuple[np.ndarray]: Prediction outputs.
        """
        preprocessed = self._preprocess(image)
        predicted = self._run_inference_client(preprocessed)
        postprocessed = self._postprocess(predicted, len(self._category_index))
        return postprocessed


class ImagePredictor(BasePredictor):
    """Generic model predictor for image tasks"""

    def __init__(
        self,
        dtype: str,
        input_path: str,
        model_name: str,
        label_map_path: str,
        threshold: float = 0.7,
        save: bool = True,
        output_path: str = "",
        **kwargs,
    ) -> None:
        super().__init__(
            dtype=dtype,
            input_path=input_path,
            model_name=model_name,
            label_map_path=label_map_path,
            threshold=threshold,
            save=save,
            output_path=output_path,
            **kwargs,
        )
        self._input_shape = [1] + self._model_config["input"][0]["dims"]
        self._input_names = [input["name"] for input in self._model_config["input"]]
        self._output_names = [output["name"] for output in self._model_config["output"]]
        self._output_shapes = [
            output["dims"] for output in self._model_config["output"]
        ]

    def _run_inference_client(self, model_input: np.ndarray) -> Dict:
        """Run inference using the Triton Inference Server client.

        Args:
            model_input (np.ndarray): The model input array.

        Returns:
            Dict: The model response outputs.
        """
        inputs = [
            httpclient.InferInput(input_name, self._input_shape, datatype="FP32")
            for input_name in self._input_names
        ]
        for inp in inputs:
            inp.set_data_from_numpy(model_input, binary_data=True)
        request_outputs = [
            httpclient.InferRequestedOutput(output_name, binary_data=True)
            for output_name in self._output_names
        ]
        response_outputs = self._triton_client.infer(
            model_name=self._model_name, inputs=inputs, outputs=request_outputs
        )
        return response_outputs


class ObjectDetectionPredictor(ImagePredictor):
    """Tensorflow Object Detection model predictor"""

    def __call__(self) -> None:
        for each_image in glob.glob(os.path.join(self._input_path, "*")):
            trt_logger.info("Predicting for %s ...", each_image)
            if not each_image.endswith((".jpg", ".jpeg", ".png")):
                continue

            scores, classes, bboxes = self._predict(each_image)

            if self._save:
                image_origi = cv2.imread(each_image)
                for box, class_id, score in zip(bboxes, classes, scores):
                    image_origi = draw_bboxes(
                        image_origi,
                        box,
                        class_id,
                        score,
                        self._color_map,
                        self._category_index,
                    )

                ## Save predicted image
                image_name = os.path.basename(each_image)
                specific_output_path = os.path.join(self._output_path, image_name)
                cv2.imwrite(specific_output_path, image_origi)
                trt_logger.info("Prediction saved to %s", specific_output_path)

    def _preprocess(self, path: str) -> np.ndarray:
        """Load and preprocess the image array before it is fed to the model.

        Args:
            path (str): The path to the image file.

        Returns:
            np.ndarray: The preprocessed image array.
        """
        image = Image.open(path).convert("RGB")
        image = image.resize((self._input_shape[1], self._input_shape[2]))
        model_input = np.array(image).astype(self._dtype)
        model_input = np.expand_dims(model_input, 0)
        return model_input

    def _postprocess(self, response_outputs, _) -> np.ndarray:
        """Postprocess the model output.

        Args:
            response_outputs (Dict): The model response outputs.
            num_classes (int): The number of classes in the model.

        Returns:
            Tuple[np.ndarray]: The postprocessed model output
                containing scores, classes, and boxes.
        """
        boxes = response_outputs.as_numpy("detection_boxes")[0]
        classes = response_outputs.as_numpy("detection_classes")[0]
        scores = response_outputs.as_numpy("detection_scores")[0]

        fltr = scores >= self._threshold
        boxes = boxes[fltr]
        classes = classes[fltr]
        scores = scores[fltr]

        if len(boxes) == 0:
            return np.array([]), np.array([]), np.array([])
        return scores, classes, boxes


class MaskPredictor(ImagePredictor):
    """Instance segmentation predictor"""

    _input_name: str = "inputs"

    def __call__(self) -> None:
        for each_image in glob.glob(os.path.join(self._input_path, "*")):
            trt_logger.info("Predicting for %s ...", each_image)
            if not each_image.endswith((".jpg", ".jpeg", ".png")):
                continue

            scores, classes, bboxes, masks = self._predict(each_image)

            if self._save:
                image_origi = cv2.imread(each_image)
                for box, mask, class_id, score in zip(bboxes, masks, classes, scores):
                    image_origi = draw_instance_masks(
                        image_origi,
                        box,
                        mask,
                        class_id,
                        self._color_map,
                        self._category_index,
                    )
                    image_origi = draw_bboxes(
                        image_origi,
                        box,
                        class_id,
                        score,
                        self._color_map,
                        self._category_index,
                    )

                ## Save predicted image
                image_name = os.path.basename(each_image)
                specific_output_path = os.path.join(self._output_path, image_name)
                cv2.imwrite(specific_output_path, image_origi)
                trt_logger.info("Prediction saved to %s", specific_output_path)

    def _preprocess(self, path: str) -> np.ndarray:
        """Load and preprocess the image array before it is fed to the model.

        Args:
            path (str): The path to the image file.

        Returns:
            np.ndarray: The preprocessed image array.
        """
        img = Image.open(os.path.join(path)).convert("RGB")
        model_input = img.resize((self._input_shape[1], self._input_shape[2]))
        model_input = np.array(model_input).astype(np.float32)
        model_input = np.expand_dims(model_input, 0)
        return model_input

    def _postprocess(self, response_outputs, _) -> np.ndarray:
        """Postprocess the model output.

        Args:
            response_outputs (Dict): The model response outputs.
            num_classes (int): The number of classes in the model.

        Returns:
            Tuple[np.ndarray]: The postprocessed model output
                containing scores, classes, boxes, and masks.
        """
        preds = []
        for output_name in self._output_names:
            output = response_outputs.as_numpy(output_name)
            preds.append(output)
        _, boxes, scores, classes, masks = preds
        scores = scores[0]
        classes = classes[0].astype(np.int16)
        boxes = boxes[0]
        masks = masks[0]
        _filter = np.where(scores > self._threshold)
        scores = scores[_filter]
        classes = classes[_filter]
        boxes = boxes[_filter]
        boxes = [
            [
                clamp(box[0]),
                clamp(box[1]),
                clamp(box[2]),
                clamp(box[3]),
            ]
            for box in boxes
        ]

        masks = masks[_filter]
        masks_output = []
        for each_mask in masks:
            instance_mask = np.zeros_like(each_mask, np.uint8)
            instance_mask[np.where(each_mask > self._threshold)] = 1
            masks_output.append(instance_mask)

        if masks_output:
            masks_output = np.stack(masks_output)
        else:
            masks_output = None
        return scores, classes, boxes, masks_output


class SemanticPredictor(ImagePredictor):
    """Semantic segmentation predictor"""

    def __call__(self) -> None:
        for each_image in glob.glob(os.path.join(self._input_path, "*")):
            trt_logger.info("Predicting for %s ...", each_image)
            if not each_image.endswith((".jpg", ".jpeg", ".png")):
                continue

            mask = self._predict(each_image)
            if mask is None:
                trt_logger.info("No detections for %s", each_image)
                continue

            if self._save:
                image_origi = cv2.imread(each_image)
                image_origi = draw_semantic_masks(
                    image_origi,
                    mask,
                    self._color_map,
                )

                ## Save predicted image
                image_name = os.path.basename(each_image)
                specific_output_path = os.path.join(self._output_path, image_name)
                cv2.imwrite(specific_output_path, image_origi)
                trt_logger.info("Prediction saved to %s", specific_output_path)

    def _preprocess(self, path: str) -> np.ndarray:
        """Load and preprocess the image array before it is fed to the model

        Args:
            path (str): The path to the image file.

        Returns:
            np.ndarray: The preprocessed image array.
        """
        img = Image.open(os.path.join(path)).convert("RGB")
        model_input = img.resize((self._input_shape[1], self._input_shape[2]))
        model_input = np.array(model_input).astype(np.float32)
        model_input = np.expand_dims(model_input, 0)
        return model_input

    def _postprocess(self, response_outputs, _) -> np.ndarray:
        """Postprocess the model output

        Args:
            response_outputs (Dict): The model response outputs.

        Returns:
            np.ndarray: The postprocessed model output
                containing the semantic mask.
        """
        mask = response_outputs.as_numpy(self._output_names[0])[0]
        binary_mask = np.zeros_like(mask[0], np.uint8)
        mask = (mask - np.min(mask)) / (np.max(mask) - np.min(mask))
        for class_id, class_mask in enumerate(mask):
            if class_id > 0:
                binary_mask[np.where(class_mask > self._threshold)] = class_id
        return binary_mask


class YOLOv8Predictor(ImagePredictor):
    """Base class for YOLOv8 predictors"""

    def _preprocess(self, path: str) -> np.ndarray:
        """Load and preprocess the image array before it is fed to the model.

        Args:
            path (str): The path to the image file.

        Returns:
            np.ndarray: The preprocessed image array.
        """
        image = Image.open(path).convert("RGB")
        image = image.resize((self._input_shape[2], self._input_shape[3]))
        model_input = np.array(image).astype(self._dtype)
        model_input = np.transpose(model_input, (2, 0, 1))
        model_input = np.expand_dims(model_input, 0)
        model_input = model_input / 255.0
        return model_input


class YOLOv8BoxPredictor(YOLOv8Predictor):
    """YOLOv8 bounding box predictor"""

    def __call__(self) -> None:
        for each_image in glob.glob(os.path.join(self._input_path, "*")):
            trt_logger.info("Predicting for %s ...", each_image)
            if not each_image.endswith((".jpg", ".jpeg", ".png")):
                continue

            scores, classes, bboxes = self._predict(each_image)

            if self._save:
                image_origi = cv2.imread(each_image)
                for box, class_id, score in zip(bboxes, classes, scores):
                    image_origi = draw_bboxes(
                        image_origi,
                        box,
                        class_id,
                        score,
                        self._color_map,
                        self._category_index,
                    )

                ## Save predicted image
                image_name = os.path.basename(each_image)
                specific_output_path = os.path.join(self._output_path, image_name)
                cv2.imwrite(specific_output_path, image_origi)
                trt_logger.info("Prediction saved to %s", specific_output_path)

    def _postprocess(self, response_outputs, num_classes: int) -> Tuple[np.ndarray]:
        """Postprocess the model output.

        Args:
            response_outputs (Dict): The model response outputs.
            num_classes (int): The number of classes in the model.

        Returns:
            Tuple[np.ndarray]: The postprocessed model output
                containing scores, classes, and boxes.
        """
        preds = []
        for output_name in self._output_names:
            output = response_outputs.as_numpy(output_name)[0]
            preds.append(output)
        prediction = preds[0]

        if len(prediction) == 0:
            return np.array([]), np.array([]), np.array([])

        mask_index = 4 + num_classes
        candidates = prediction[4:mask_index].max(0) > 0.05
        prediction = prediction.transpose()[candidates]

        scores, classes, boxes, _ = postprocess_predictions(prediction, num_classes)

        if len(boxes) > 0:
            boxes, classes, scores, _, _, _ = nms_boxes(
                boxes, classes, scores, 0.7, self._dtype
            )
            boxes = xywh2xyxy(boxes, self._input_shape[2], self._input_shape[3])
        fltr = scores > self._threshold
        scores = scores[fltr]
        boxes = boxes[fltr]
        classes = classes[fltr]
        boxes = [
            [
                clamp(box[1]),
                clamp(box[0]),
                clamp(box[3]),
                clamp(box[2]),
            ]
            for box in boxes
        ]  # yxyx
        return scores, classes, boxes


class YOLOv8MaskPredictor(YOLOv8Predictor):
    """YOLOv8 instance mask predictor"""

    def __call__(self) -> None:
        for each_image in glob.glob(os.path.join(self._input_path, "*")):
            trt_logger.info("Predicting for %s ...", each_image)
            if not each_image.endswith((".jpg", ".jpeg", ".png")):
                continue

            scores, classes, bboxes, masks = self._predict(each_image)

            if self._save:
                image_origi = cv2.imread(each_image)
                for box, mask, class_id, score in zip(bboxes, masks, classes, scores):
                    image_origi = draw_yolov8_instance_masks(
                        image_origi,
                        box,
                        mask,
                        class_id,
                        self._color_map,
                        self._category_index,
                    )
                    image_origi = draw_bboxes(
                        image_origi,
                        box,
                        class_id,
                        score,
                        self._color_map,
                        self._category_index,
                    )

                ## Save predicted image
                image_name = os.path.basename(each_image)
                specific_output_path = os.path.join(self._output_path, image_name)
                cv2.imwrite(specific_output_path, image_origi)
                trt_logger.info("Prediction saved to %s", specific_output_path)

    def _postprocess(self, response_outputs, num_classes: int) -> Tuple[np.ndarray]:
        """Postprocess the model output.

        Args:
            response_outputs (Dict): The model response outputs.
            num_classes (int): The number of classes in the model.

        Returns:
            Tuple[np.ndarray]: The postprocessed model output containing
                scores, classes, boxes, and masks.
        """
        preds = []
        for output_name in self._output_names:
            output = response_outputs.as_numpy(output_name)
            preds.append(output)

        protos = preds[0]
        predictions = preds[1]

        # Take only the first batch
        proto = protos[0]
        prediction = predictions[0]

        mask_index = 4 + num_classes
        candidates = prediction[4:mask_index].max(0) > 0.05
        prediction = prediction.transpose()[candidates]

        if len(prediction) == 0:
            return (
                np.ndarray(0, dtype=np.int32),
                np.ndarray(0, dtype=np.float32),
                np.ndarray((0, 4), dtype=np.float32),
                np.ndarray(0, dtype=np.float32),
            )

        scores, classes, boxes, masks = postprocess_predictions(prediction, num_classes)

        # Perform the below only for yolov8seg models.
        channels, mask_height, mask_width = proto.shape
        masks = masks @ proto.astype(np.float64).reshape(channels, -1)
        masks = 1 / (1 + np.exp(-1 * masks))  # use -1*mask over -mask to avoid overflow

        masks = masks.reshape(-1, mask_height, mask_width)
        binary_masks = np.where(masks > 0.1, np.uint8(1), np.uint8(0))

        if len(boxes) > 0:
            boxes, classes, scores, binary_masks, _, _ = nms_boxes(
                boxes, classes, scores, 0.7, self._dtype, binary_masks
            )
            boxes = xywh2xyxy(boxes, self._input_shape[2], self._input_shape[3])

        # Filter out low confidence predictions and prepare boxes for drawing
        fltr = scores > self._threshold
        scores = scores[fltr]
        boxes = boxes[fltr]
        classes = classes[fltr]
        binary_masks = binary_masks[fltr]
        boxes = [
            [
                clamp(box[1]),
                clamp(box[0]),
                clamp(box[3]),
                clamp(box[2]),
            ]
            for box in boxes
        ]  # y1, x1, y2, x2

        return scores, classes, boxes, binary_masks


class YOLOv8PosePredictor(YOLOv8Predictor):
    """YOLOv8 pose predictor"""

    def __call__(self) -> None:
        skeletons = load_skeleton(self._skeleton_file_path)

        for each_image in glob.glob(os.path.join(self._input_path, "*")):
            trt_logger.info("Predicting for %s ...", each_image)
            if not each_image.endswith((".jpg", ".jpeg", ".png")):
                continue

            _, classes, _, keypoints, keypoint_confidences = self._predict(each_image)

            if self._save:
                image_origi = cv2.imread(each_image)
                for keypoint, keypoint_confidence, class_id in zip(
                    keypoints, keypoint_confidences, classes
                ):
                    image_origi = draw_keypoints(
                        image_origi,
                        keypoint,
                        keypoint_confidence,
                        class_id,
                        self._color_map,
                        self._category_index,
                        skeletons,
                    )

                ## Save predicted image
                image_name = os.path.basename(each_image)
                specific_output_path = os.path.join(self._output_path, image_name)
                cv2.imwrite(specific_output_path, image_origi)
                trt_logger.info("Prediction saved to %s", specific_output_path)

    def _postprocess(self, response_outputs, num_classes: int) -> Tuple[np.ndarray]:
        """Postprocess the model output.

        Args:
            response_outputs (Dict): The model response outputs.
            num_classes (int): The number of classes in the model.

        Returns:
            Tuple[np.ndarray]: The postprocessed model output containing
                scores, classes, boxes, keypoints, and keypoint confidences.
        """
        preds = []
        for output_name in self._output_names:
            output = response_outputs.as_numpy(output_name)[0]
            preds.append(output)

        # Take only the first batch
        prediction = preds[0]

        keypoints_index = 4 + num_classes
        candidates = prediction[4:keypoints_index].max(0) > 0.05

        prediction = prediction.transpose()[candidates]

        if len(prediction) == 0:
            return (
                np.ndarray(0, dtype=np.int32),
                np.ndarray(0, dtype=np.float32),
                np.ndarray((0, 4), dtype=np.float32),
                np.ndarray((0, 3), dtype=np.float32),
                np.ndarray((0, 3), dtype=np.float32),
            )

        scores, classes, boxes, keypoints = postprocess_predictions(
            prediction, num_classes
        )

        # Perform the below only for yolov8pose models
        keypoints = keypoints.reshape(len(scores), -1, 3)
        keypoints[..., 0] /= self._input_shape[3]
        keypoints[..., 1] /= self._input_shape[2]

        keypoint_confidences = keypoints[..., 2].copy()

        keypoints[..., 2] = np.where(
            np.any(keypoints[..., :2] > 1.0, axis=-1)
            | np.any(keypoints[..., :2] < 0.0, axis=-1),
            0,
            2,
        )

        if len(boxes) > 0:
            boxes, classes, scores, _, keypoints, keypoint_confidences = nms_boxes(
                boxes,
                classes,
                scores,
                0.7,
                self._dtype,
                keypoints=keypoints,
                keypoint_confs=keypoint_confidences,
            )
            boxes = xywh2xyxy(boxes, self._input_shape[2], self._input_shape[3])

        # Filter out low confidence predictions and prepare boxes for drawing
        fltr = scores > self._threshold
        scores = scores[fltr]
        classes = classes[fltr]
        boxes = boxes[fltr]
        keypoints = keypoints[fltr]
        keypoint_confidences = keypoint_confidences[fltr]

        return scores, classes, boxes, keypoints, keypoint_confidences


class YOLOv8ClassificationPredictor(YOLOv8Predictor):
    """YOLOv8 classification predictor"""

    def __call__(self) -> None:
        for each_image in glob.glob(os.path.join(self._input_path, "*")):
            trt_logger.info("Predicting for %s ...", each_image)
            if not each_image.endswith((".jpg", ".jpeg", ".png")):
                continue

            scores, classes = self._predict(each_image)

            if self._save:
                image_origi = cv2.imread(each_image)
                for class_id, score in zip(classes, scores):
                    image_origi = draw_class_labels(
                        image_origi,
                        class_id,
                        score,
                        self._color_map,
                        self._category_index,
                    )
                ## Save predicted image
                image_name = os.path.basename(each_image)
                specific_output_path = os.path.join(self._output_path, image_name)
                cv2.imwrite(specific_output_path, image_origi)
                trt_logger.info("Prediction saved to %s", specific_output_path)

    def _postprocess(self, response_outputs, _) -> Tuple[np.ndarray]:
        """Postprocess the model output.

        Args:
            response_outputs (Dict): The model response outputs.

        Returns:
            Tuple[np.ndarray]: The postprocessed model output containing
                scores and classes.
        """
        preds = []
        for output_name in self._output_names:
            output = response_outputs.as_numpy(output_name)[0]
            preds.append(output)

        pred = preds[0]
        score = np.max(pred)
        cls = np.argmax(pred)
        score = np.expand_dims(score, axis=0).astype(np.float32)
        cls = np.expand_dims(cls, axis=0).astype(np.int32)
        return score, cls


class VideoClassificationPredictor(BasePredictor):
    """Video classification predictor"""

    def __init__(
        self,
        dtype: str,
        input_path: str,
        model_name: str,
        label_map_path: str,
        threshold: float = 0.7,
        save: bool = True,
        output_path: str = "",
        **kwargs,
    ) -> None:
        super().__init__(
            dtype=dtype,
            input_path=input_path,
            model_name=model_name,
            label_map_path=label_map_path,
            threshold=threshold,
            save=save,
            output_path=output_path,
            **kwargs,
        )
        self._states = []
        self._input_index = 0
        for i, item in enumerate(self._model_config["input"]):
            if item["name"] == "image":
                input_state = httpclient.InferInput(
                    item["name"],
                    [1] + item["dims"],
                    datatype="FP32" if item["data_type"] == "TYPE_FP32" else "INT32",
                )
                self._input_shape = item["dims"]
                self._input_index = i
            else:
                input_state = httpclient.InferInput(
                    item["name"],
                    [1] + item["dims"],
                    datatype="FP32" if item["data_type"] == "TYPE_FP32" else "INT32",
                )
                input_state.set_data_from_numpy(
                    (
                        np.zeros(input_state.shape(), dtype=np.float32)
                        if item["data_type"] == "TYPE_FP32"
                        else np.zeros(input_state.shape(), dtype=np.int32)
                    ),
                    binary_data=True,
                )
            self._states.append(input_state)
        self._outputs = [
            httpclient.InferRequestedOutput(output["name"], binary_data=True)
            for output in self._model_config["output"]
        ]
        self._output_name = "output_1"

    def __call__(self) -> None:
        for each_video in glob.glob(os.path.join(self._input_path, "*")):
            trt_logger.info("Predicting for %s ...", each_video)
            pred_label, conf = self._predict(each_video)
            trt_logger.info("Prediction: %s", pred_label)
            trt_logger.info("Confidence score: %s", conf)

    def _run_inference_client(self, model_input: np.ndarray) -> Dict:
        """Run inference using the Triton Inference Server client.

        Args:
            model_input (np.ndarray): The model input array.

        Returns:
            Dict: The model response outputs.
        """
        all_logits = None
        for clip in model_input:
            self._states[self._input_index].set_data_from_numpy(
                clip.astype(np.float32), binary_data=True
            )

            response_outputs = self._triton_client.infer(
                model_name=self._model_name,
                inputs=self._states,
                outputs=self._outputs,
            )

            for state in self._states:
                result = response_outputs.as_numpy(state.name())
                if result is not None:
                    state.set_data_from_numpy(
                        result,
                        binary_data=True,
                    )
            logits = response_outputs.as_numpy(self._output_name)
            all_logits = logits
        return all_logits

    def _preprocess(self, path: str) -> np.ndarray:
        """Load video and sample frames.

        Args:
            path (str): The path to the video file.

        Returns:
            np.ndarray: The preprocessed video array.
        """
        video = cv2.VideoCapture(path)
        frames = []
        video_length = video.get(cv2.CAP_PROP_FRAME_COUNT)
        for _ in range(int(video_length)):
            ret, frame = video.read()
            if not ret:
                break
            frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
            frame = cv2.resize(frame, (self._input_shape[1], self._input_shape[2]))
            frame = np.expand_dims(np.expand_dims(frame, 0), 0)
            frames.append(frame)
        clips = np.array(frames, dtype=np.float32)
        return clips

    def _postprocess(self, response_outputs, _) -> np.ndarray:
        """Postprocess the model output.

        Args:
            response_outputs (Dict): The model response outputs.
            num_classes (int): The number of classes in the model.

        Returns:
            Tuple: The postprocessed model output
                containing class name and confidence score.
        """
        probs = response_outputs.flatten()
        label = np.argmax(probs)
        score = max(probs)
        label = self._category_index[label]["name"]
        return label, score



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   tag.py
@Author  :   Raighne.Weng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Datature Tag API
"""

from typing import Union

from datature.nexus import models
from datature.nexus.api.types import TagMetadata
from datature.nexus.client_context import RestContext


class Tag(RestContext):
    """Datature Tag API Resource."""

    def list(self) -> models.Tags:
        """Lists all tags in the project.

        :return: A msgspec struct containing tag metadata with the following structure:

            .. code-block:: python

                [Tag(
                    index=0,
                    name='Platelets',
                    color=None,
                    description=None
                )]

        :example:
                .. code-block:: python

                        from datature.nexus import Client

                        project = Client("5aa41e8ba........").get_project("proj_b705a........")
                        project.tags.list()
        """
        return self.requester.GET(
            f"/projects/{self.project_id}/tags", response_type=models.Tags
        )

    def create(self, metadata: Union[TagMetadata, dict]) -> models.Tags:
        """Creates a new tag.
            The indices of new tags created will begin from the last existing tag index.

        :param metadata: The metadata of the new tag.
        :return: An updated msgspec struct containing tag metadata with the following structure:

            .. code-block:: python

                [
                    Tag(
                        index=0,
                        name='boat',
                        color='#7ed957',
                        description='boat'
                    ),
                    Tag(
                        index=1,
                        name='boat2',
                        color='#ff3131',
                        description=None
                    )
                ]


        :example:
            .. code-block:: python

                from datature.nexus import Client, ApiTypes

                project = Client("5aa41e8ba........").get_project("proj_b705a........")

                project.tags.create({"name": "boat2", "color": "#ff3131"}

                # or
                project.tags.create(ApiTypes.TagMetadata(
                                        name="boat2",
                                        color="#ff3131",
                                        description="This is a boat"
                                    ))
        """
        assert isinstance(metadata, (TagMetadata, dict))

        if isinstance(metadata, dict):
            metadata = TagMetadata(**metadata)

        return self.requester.POST(
            f"/projects/{self.project_id}/tags",
            request_body=metadata.to_json(),
            response_type=models.Tags,
        )

    def update(self, index: int, metadata: Union[TagMetadata, dict]) -> models.Tags:
        """Updates the name of a specific tag using the tag index.

        :param index: The index of the tag to update.
        :param metadata: The new metadata of the tag.
        :return: An updated msgspec struct containing tag metadata with the following structure:

            .. code-block:: python

                [
                    Tag(
                        index=0,
                        name='boat',
                        color='#7ed957',
                        description='boat'
                    ),
                    Tag(
                        index=1,
                        name='tagName',
                        color='#ff3131',
                        description=None
                    )
                ]

        :example:
                .. code-block:: python

                        from datature.nexus import Client

                        project = Client("5aa41e8ba........").get_project("proj_b705a........")

                        project.tags.update(1, "tagName")
        """
        assert isinstance(index, int)
        assert isinstance(metadata, (TagMetadata, dict))

        if isinstance(metadata, dict):
            metadata = TagMetadata(**metadata)

        return self.requester.PATCH(
            f"/projects/{self.project_id}/tags/{index}",
            request_body=metadata.to_json(),
            response_type=models.Tags,
        )

    def delete(self, index: int) -> models.DeleteResponse:
        """Deletes a specific tag using the tag index.
            The tag indices of other tags will be left unchanged.
            The indices of new tags created will begin from the last existing tag index.

        :param index: The index of the tag.
        :return: A msgspec struct containing the deletion
            status of the tag with the following structure:

                .. code-block:: python

                    DeleteResponse(deleted=True, id=None)

        :example:
            .. code-block:: python

                from datature.nexus import Client

                project = Client("5aa41e8ba........").get_project("proj_b705a........")

                project.tags.delete(1)
        """
        assert isinstance(index, int)
        return self.requester.DELETE(
            f"/projects/{self.project_id}/tags/{index}",
            response_type=models.DeleteResponse,
        )

    def merge(self, index: int, target_index: int) -> models.Tags:
        """Merges a tag into another tag. .

        :param index: The index of the tag to merge.
        :param target_index: The index of the tag to merge into.
        :return: An updated msgspec struct containing tag
            metadata with the following structure:

            .. code-block:: python

                [Tag(
                    index=0,
                    name='boat',
                    color='#7ed957',
                    description='boat'
                )]

        :example:
            .. code-block:: python

                from datature.nexus import Client

                project = Client("5aa41e8ba........").get_project("proj_b705a........")
                project.tags.merge(1, 0)
        """
        assert isinstance(index, int)
        assert isinstance(target_index, int)

        return self.requester.POST(
            f"/projects/{self.project_id}/tags/{index}-{target_index}:merge",
            response_type=models.Tags,
        )



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   workflow.py
@Author  :   Raighne.Weng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Datature Workflow API
"""

from typing import Union

from datature.nexus import models
from datature.nexus.api.types import FlowMetadata
from datature.nexus.client_context import RestContext


class Workflow(RestContext):
    """Datature Workflow API Resource."""

    def list(self) -> models.Workflows:
        """Lists all workflows in the project.

        :return:
            A list of msgspec structs containing the workflow metadata with the following structure:

                .. code-block:: python

                    [Workflow(
                        id='flow_64e812a7e47592ef374cbbc2',
                        project_id='proj_cd067221d5a6e4007ccbb4afb5966535',
                        title='Yolov8 Workflow',
                        create_date=1701927649302,
                        update_date=1701927649302
                    )]

        :example:
                .. code-block:: python

                    from datature.nexus import Client

                    project = Client("5aa41e8ba........").get_project("proj_b705a........")
                    project.workflows.list()
        """
        return self.requester.GET(
            f"/projects/{self.project_id}/workflows", response_type=models.Workflows
        )

    def get(self, flow_id: str) -> models.Workflow:
        """Retrieves a specific workflow using the flow ID.

        :param flow_id: The ID of the workflow.
        :return:
            A msgspec struct containing the specific workflow metadata with the following structure:

                .. code-block:: python

                    Workflow(
                        id='flow_64e812a7e47592ef374cbbc2',
                        project_id='proj_cd067221d5a6e4007ccbb4afb5966535',
                        title='Yolov8 Workflow',
                        create_date=1701927649302,
                        update_date=1701927649302
                    )


        :example:
                .. code-block:: python

                        from datature.nexus import Client

                        project = Client("5aa41e8ba........").get_project("proj_b705a........")
                        project.workflows.get("flow_639309be08b4488a914b8802")
        """
        assert isinstance(flow_id, str)

        return self.requester.GET(
            f"/projects/{self.project_id}/workflows/{flow_id}",
            response_type=models.Workflow,
        )

    def update(self, flow_id: str, flow: Union[FlowMetadata, dict]) -> models.Workflow:
        """Updates title of a specific workflow using the flow ID.

        :param flow_id: The ID of the workflow.
        :param flow: The new metadata of the workflow to be updated.
        :return: A msgspec struct containing the updated workflow metadata with the following structure:

                .. code-block:: python

                    Workflow(
                        id='flow_64e812a7e47592ef374cbbc2',
                        project_id='proj_cd067221d5a6e4007ccbb4afb5966535',
                        title='My awesome workflow',
                        create_date=1701927649302,
                        update_date=1701927649302
                    )

        :example:
                .. code-block:: python

                        from datature.nexus import Client, ApiTypes

                        project = Client("5aa41e8ba........").get_project("proj_b705a........")

                        project.workflows.update(
                            "flow_639309be08b4488a914b8802",
                            {"title": "My awesome workflow"}
                        )
                        // or
                        project.workflows.update(
                            "flow_639309be08b4488a914b8802",
                            ApiTypes.FlowMetadata(title="My awesome workflow")
                        )
        """
        assert isinstance(flow_id, str)
        assert isinstance(flow, (FlowMetadata, dict))

        if isinstance(flow, dict):
            flow = FlowMetadata(**flow)

        return self.requester.PATCH(
            f"/projects/{self.project_id}/workflows/{flow_id}",
            request_body=flow.to_json(),
            response_type=models.Workflow,
        )

    def delete(self, flow_id: str) -> models.DeleteResponse:
        """Deletes a specific workflow using the flow ID.

        :param flow_id: The ID of the workflow
        :return: A msgspec struct containing the deleted
            workflow ID and deletion status with the following structure:

            .. code-block:: python

                DeleteResponse(deleted=True, id='flow_656d7912092f3395248c7666')

        :example:
            .. code-block:: python

                from datature.nexus import Client

                project = Client("5aa41e8ba........").get_project("proj_b705a........")
                project.workflows.delete("flow_639309be08b4488a914b8802")
        """
        assert isinstance(flow_id, str)

        return self.requester.DELETE(
            f"/projects/{self.project_id}/workflows/{flow_id}",
            response_type=models.DeleteResponse,
        )



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   operation.py
@Author  :   Raighne.Weng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Datature Operation API
"""

import logging
import time
from datetime import datetime, timedelta
from typing import Union

from datature.nexus import config, error, models
from datature.nexus.api.types import OperationStatusOverview
from datature.nexus.client_context import RestContext

logger = logging.getLogger("datature-nexus")


class Operation(RestContext):
    """Datature Operation API Resource."""

    def get(self, op_id: str) -> models.Operation:
        """Retrieves an executed operation status using the operation link.

        :param op_id: The id of the operation as a string.
        :return: A msgspec struct containing the operation metadata with the following structure.

                .. code-block:: python

                        Operation(
                            id='op_d52ab4e6-7760-4d12-9c43-7fc1b8ce1616',
                            kind='nexus.annotations.export',
                            status=OperationStatus(
                                overview='Finished',
                                message='Operation finished',
                                progress=OperationProgress(
                                    unit='whole operation',
                                    with_status=OperationProgressWithStatus(
                                        Queued=0,
                                        Running=0,
                                        Finished=1,
                                        Cancelled=0,
                                        Errored=0
                                    )
                                )
                            ),
                            create_date=1701927649302,
                            update_date=1701927649302
                        )

        :example:
                .. code-block:: python

                        from datature.nexus import Client

                        project = Client("5aa41e8ba........").get_project("proj_b705a........")

                        project.operations.get("op_508fc5d1-e908-486d-9e7b-1dca99b80024")
        """
        assert isinstance(op_id, str)
        return self.requester.GET(
            f"/projects/{self.project_id}/operations/{op_id}",
            response_type=models.Operation,
        )

    def wait_until_done(
        self,
        op_id: str,
        timeout: int = config.OPERATION_LOOPING_TIMEOUT_SECONDS,
        raise_exception_if: Union[
            OperationStatusOverview, str
        ] = OperationStatusOverview.ERRORED,
    ) -> models.Operation:
        """Continuously retrieves an executed operation status for a
        specified number of times at a specified delay interval.

        Can be used to poll a running operation to monitor execution status.

        :param op_id: The id of the operation as a string.
        :param timeout: The maximum number of times to loop the operation retrieval.
        :param raise_exception_if: The condition to raise error.
        :return: The operation status metadata if the operation has finished,
            a BadRequestError if the operation has errored out,
            or None if the operation is still running.

                .. code-block:: python

                    Operation(
                        id='op_d52ab4e6-7760-4d12-9c43-7fc1b8ce1616',
                        kind='nexus.annotations.export',
                        status=OperationStatus(
                            overview='Finished',
                            message='Operation finished',
                            progress=OperationProgress(
                                unit='whole operation',
                                with_status=OperationProgressWithStatus(
                                    Queued=0,
                                    Running=0,
                                    Finished=1,
                                    Cancelled=0,
                                    Errored=0
                                )
                            )
                        ),
                        create_date=1701927649302,
                        update_date=1701927649302
                    )

        :example:
                .. code-block:: python

                        from datature.nexus import Client

                        project = Client("5aa41e8ba........").get_project("proj_b705a........")

                        project.operations.wait_until_done(
                            "op_508fc5d1-e908-486d-9e7b-1dca99b80024"
                        )
        """
        assert isinstance(op_id, str)
        assert isinstance(raise_exception_if, (str, OperationStatusOverview))

        if isinstance(raise_exception_if, str):
            raise_exception_if = OperationStatusOverview(raise_exception_if)

        timeout = timeout or config.OPERATION_LOOPING_TIMEOUT_SECONDS
        response = None
        elapsed_time = datetime.now() + timedelta(seconds=timeout)

        while True:
            response = self.requester.GET(
                f"/projects/{self.project_id}/operations/{op_id}",
                response_type=models.Operation,
            )

            logger.debug("Operation status: %s ", response.status)

            if response.status.overview == OperationStatusOverview.FINISHED.value:
                return response

            if response.status.overview == raise_exception_if.value:
                logger.warning(
                    "Operation %s %s: please contact support at support@datature.io",
                    raise_exception_if.value,
                    op_id,
                )

                raise error.BadRequestError(
                    f"Operation {raise_exception_if.value} {op_id}: "
                    "please contact support at support@datature.io"
                )

            # if the operation has not finished when the timeouts
            if elapsed_time < datetime.now():
                logger.warning(
                    "Operation timeout: please use operation.get(%s) to get the status",
                    op_id,
                )
                return response

            time.sleep(config.OPERATION_LOOPING_DELAY_SECONDS)



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   test_annotation.py
@Author  :   Raighne.Weng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Datature Annotation API Test Cases
"""

import json
import unittest
from unittest.mock import ANY, MagicMock, patch

import msgspec

from datature.nexus import models
from datature.nexus.api.annotation.annotation import Annotation
from datature.nexus.api.annotation.import_session import ImportSession
from datature.nexus.client_context import ClientContext
from tests.unit.fixture.data import operation_fixture
from tests.unit.fixture.mock import MockResponse

# pylint: disable=W0212,W0702


class TestAnnotation(unittest.TestCase):
    """Datature Annotation API Resource Test Cases."""

    def __init__(self, *args, **kwargs):
        """init global variables."""
        unittest.TestCase.__init__(self, *args, **kwargs)

        context = ClientContext("secret_key", endpoint="http://localhost:8080")
        context.project_id = "project_id"
        context.requester._request = MagicMock()

        self.annotation = Annotation(context)
        # Mock operation
        self.annotation.operation.wait_until_done = MagicMock()

    def test_list(self):
        """Test retrieve a list of annotations."""
        self.annotation.list()

        self.annotation._context.requester._request.assert_called_once_with(
            "GET",
            "/projects/project_id/annotations",
            ANY,
            query={"limit": 100, "includeAttributes": False},
            request_params=None,
        )

    def test_create(self):
        """Test create an annotation."""
        self.annotation.create(
            {
                "asset_id": "asset_id",
                "tag": "tagName",
                "bound_type": "Rectangle",
                "bound": [
                    [0.425, 0.49382716049382713],
                    [0.425, 0.6419753086419753],
                    [0.6, 0.6419753086419753],
                    [0.6, 0.49382716049382713],
                ],
            }
        )

        self.annotation._context.requester._request.assert_called_once_with(
            "POST",
            "/projects/project_id/annotations",
            ANY,
            request_body={
                "assetId": "asset_id",
                "tag": "tagName",
                "boundType": "Rectangle",
                "bound": [
                    [0.425, 0.49382716049382713],
                    [0.425, 0.6419753086419753],
                    [0.6, 0.6419753086419753],
                    [0.6, 0.49382716049382713],
                ],
            },
            query=None,
            request_params=None,
        )

    def test_get(self):
        """Test retrieve an annotation."""
        self.annotation.get("annotation_id")

        self.annotation._context.requester._request.assert_called_once_with(
            "GET",
            "/projects/project_id/annotations/annotation_id",
            ANY,
            query={"includeAttributes": False},
            request_params=None,
        )

    def test_delete(self):
        """Test delete an annotation."""
        self.annotation.delete("annotation_id")

        self.annotation._context.requester._request.assert_called_once_with(
            "DELETE",
            "/projects/project_id/annotations/annotation_id",
            ANY,
        )

    def test_export_with_background(self):
        """Test export annotations."""
        self.annotation.create_export(
            {
                "format": "csv_fourcorner",
                "options": {"split_ratio": 0.4, "seed": 1337, "normalized": True},
            },
            background=True,
        )

        self.annotation._context.requester._request.assert_called_once_with(
            "POST",
            "/projects/project_id/annotationexports",
            ANY,
            query=None,
            request_body={
                "format": "csv_fourcorner",
                "options": {
                    "splitRatio": 0.4,
                    "seed": 1337,
                    "normalized": True,
                },
            },
            request_params=None,
        )

    def test_export(self):
        """Test export annotations."""
        self.annotation.create_export(
            {
                "format": "csv_fourcorner",
                "options": {
                    "split_ratio": 0.4,
                    "seed": 1337,
                },
            }
        )
        self.annotation._context.requester._request.side_effect = MockResponse(
            operation_fixture.pending_operation_response, 200
        )

        self.annotation.operation.wait_until_done.assert_called()

    @patch("datature.nexus.api.annotation.annotation.utils")
    @patch("datature.nexus.api.annotation.annotation.zipfile")
    def test_download_exported_file(self, patch_utils, patch_zipfile):
        """Test get export annotations."""
        patch_utils.download_files_to_tempfile = MagicMock()
        patch_zipfile.ZipFile = MagicMock()

        self.annotation._context.requester._request.return_value = msgspec.json.decode(
            json.dumps(operation_fixture.annotation_operation_response),
            type=models.ExportedAnnotations,
        )

        self.annotation.download_exported_file("op_id")

    def test_download_exported_file_error(self):
        """Test get export annotations."""
        self.annotation._context.requester._request.return_value = msgspec.json.decode(
            json.dumps(operation_fixture.annotation_export_error_response),
            type=models.ExportedAnnotations,
        )
        try:
            self.annotation.download_exported_file("op_id")
        except:
            pass

    def test_create_import_session(self):
        """Test export annotations."""
        import_session = self.annotation.create_import_session()

        assert isinstance(import_session, ImportSession)

    def test_list_import_sessions(self):
        """Test export annotations."""
        self.annotation.list_import_sessions()

        self.annotation._context.requester._request.assert_called_once_with(
            "GET",
            "/projects/project_id/annotationimportsessions",
            ANY,
            query=None,
            request_params=None,
        )

    def test_get_import_session(self):
        """Test export annotations."""
        self.annotation.get_import_session("import_session_id")

        self.annotation._context.requester._request.assert_called_once_with(
            "GET",
            "/projects/project_id/annotationimportsessions/import_session_id",
            ANY,
            query=None,
            request_params=None,
        )

    def test_get_import_session_logs(self):
        """Test export annotations."""
        self.annotation.get_import_session_logs("import_session_id")

        self.annotation._context.requester._request.assert_called_once_with(
            "GET",
            "/projects/project_id/annotationimportsessions/import_session_id/logs",
            ANY,
            query=None,
            request_params=None,
        )



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   annotation.py
@Author  :   Raighne.Weng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Datature Annotation API
"""

import json
import zipfile
from pathlib import Path
from typing import Union

import msgspec

from datature.nexus import error, models
from datature.nexus.api.annotation.import_session import ImportSession
from datature.nexus.api.operation import Operation
from datature.nexus.api.types import (
    AnnotationExportMetadata,
    AnnotationFilter,
    AnnotationMetadata,
    Pagination,
)
from datature.nexus.client_context import ClientContext, RestContext
from datature.nexus.utils import utils


class Annotation(RestContext):
    """Datature Annotation API Resource."""

    def __init__(self, client_context: ClientContext):
        """Initialize the API Resource."""
        super().__init__(client_context)
        self.operation = Operation(client_context)

    def list(
        self,
        pagination: Union[Pagination, dict, None] = None,
        filters: Union[AnnotationFilter, dict, None] = None,
        include_attributes: bool = False,
    ) -> models.PaginationResponse[models.Annotation]:
        """Lists all annotations of a specific asset.

        :param pagination: The pagination options.
        :param filters: The filter options.
        :param include_attributes: A flag indicating whether to include the annotation attributes.
        :return: A list of msgspec struct containing
            annotation metadata with the following structure:

            .. code-block:: python

                PaginationResponse(
                    next_page='NjVhYTE0ZDY3YzQ5MThlMTJjYTUyOGRk',
                    prev_page=None,
                    data=[
                        Annotation(
                            id='annot_e5028bdc-122e-4837-b354-d82b95320b69',
                            project_id='proj_cd067221d5a6e4007ccbb4afb5966535',
                            asset_id='asset_6647b58e-86af-460c-9a85-efead830aee8',
                            tag='dog',
                            bound_type='Polygon',
                            bound=[
                                [0.42918, 0.45322666666666667],
                                [0.43677999999999995, 0.4579466666666666],
                                [0.43677999999999995, 0.46941333333333335],
                                [0.44236000000000003, 0.4788800000000001],
                                [0.45146, 0.4829333333333333],
                                [0.45754, 0.48696000000000006],
                                [0.48234, 0.4842666666666666],
                                [0.4869, 0.43432000000000004],
                                [0.50514, 0.4302666666666667]
                            ],
                            create_date=None,
                            ontology_id='ontology_843e486c-58d7-45a7-b722-f4948e204a56',
                            attributes=[
                                OntologyAttributeValue(
                                    id='attri_d0877827-63d6-4794-b520-ef3e0c57ef71',
                                    name='Tag Group',
                                    value=['person']
                                )
                            ]
                        )
                    ]
                )

        :example:
            .. code-block:: python

                from datature.nexus import Client

                project = Client("5aa41e8ba........").get_project("proj_b705a........")
                project.annotations.list(
                    pagination = {
                        "limit": 500,
                        "page": "NjVhYTE0ZDY3YzQ5MThlMTJjYTUyOGRk"
                    },
                    filters = {
                        "asset_ids": ["asset_6647b58e-86af-460c-9a85-efead830aee8"]
                    }
                )
        """
        assert isinstance(pagination, (Pagination, dict, type(None)))
        assert isinstance(filters, (AnnotationFilter, dict, type(None)))

        if isinstance(pagination, dict):
            pagination = Pagination(**pagination)
        if pagination is None:
            pagination = Pagination()

        if isinstance(filters, dict):
            filters = AnnotationFilter(**filters)
        if filters is None:
            filters = AnnotationFilter()

        return self.requester.GET(
            f"/projects/{self.project_id}/annotations",
            query={
                **pagination.to_json(),
                **filters.to_json(),
                "includeAttributes": include_attributes,
            },
            response_type=models.PaginationResponse[models.Annotation],
        )

    def create(self, annotation: Union[AnnotationMetadata, dict]) -> models.Annotation:
        """Creates an annotation.

        :param annotation: The metadata of the annotation.
        :return: A msgspec struct containing the annotation metadata with the following structure:

            .. code-block:: python

                Annotation(
                    id='annot_ffcd02f1-bac2-4cd4-b959-a04e0c67b44e',
                    project_id='proj_cd067221d5a6e4007ccbb4afb5966535',
                    asset_id='asset_dd0c72a4-a8c4-479e-a130-432e699351fc',
                    tag='boat',
                    bound_type='Rectangle',
                    bound=[
                        [0.425, 0.49382716049382713],
                        [0.425, 0.6419753086419753],
                        [0.6, 0.6419753086419753],
                        [0.6, 0.49382716049382713]
                    ],
                    create_date=1701927649302,
                )

        :example:
            .. code-block:: python

                from datature.nexus import Client

                project = Client("5aa41e8ba........").get_project("proj_b705a........")

                project.annotations.create({
                    "bound_type": "Rectangle",
                    "bound": [[0.425, 0.49382716049382713], [0.425, 0.6419753086419753],
                            [0.6, 0.6419753086419753], [0.6, 0.49382716049382713]],
                    "asset_id":
                        "asset_6aea3395-9a72-4bb5-9ee0-19248c903c56",
                    "tag":
                        "tagName1"
                })
        """
        assert isinstance(annotation, (AnnotationMetadata, dict))

        if isinstance(annotation, dict):
            annotation = AnnotationMetadata(**annotation)

        return self.requester.POST(
            f"/projects/{self.project_id}/annotations",
            request_body=annotation.to_json(),
            response_type=models.Annotation,
        )

    def get(
        self, annotation_id: str, include_attributes: bool = False
    ) -> models.Annotation:
        """Retrieves a specific annotation using the annotation ID.

        :param annotation_id: The ID of the annotation.
        :param include_attributes: A flag indicating whether to include the annotation attributes.
        :return: A msgspec struct containing
            the specific annotation metadata with the following structure:

            .. code-block:: python

                    Annotation(
                        id='annot_a9ff9b21-c0e2-49ff-8a69-773aaf00a6f8',
                        project_id='proj_cd067221d5a6e4007ccbb4afb5966535',
                        asset_id='asset_f4dcb429-0332-4dd6-a1b4-fee794031ba6',
                        tag='boat',
                        bound_type='Rectangle',
                        create_date=1701927649302,
                        bound=[
                            [0.2772511848341232, 0.34635416666666663],
                            [0.2772511848341232, 0.46875],
                            [0.54739336492891, 0.46875],
                            [0.54739336492891, 0.34635416666666663]
                        ],
                        ontology_id='ontology_843e486c-58d7-45a7-b722-f4948e204a56',
                        attributes=[
                            OntologyAttributeValue(
                                id='attri_d0877827-63d6-4794-b520-ef3e0c57ef71',
                                name='Tag Group',
                                value=['person']
                            )
                        ]
                    )

        :example:
            .. code-block:: python

                    from datature.nexus import Client

                    project = Client("5aa41e8ba........").get_project("proj_b705a........")
                    project.annotations.get(
                        "asset_6aea3395-9a72-4bb5-9ee0-19248c903c56",
                        include_attributes=True
                    )
        """
        assert isinstance(annotation_id, str)
        return self.requester.GET(
            f"/projects/{self.project_id}/annotations/{annotation_id}",
            query={"includeAttributes": include_attributes},
            response_type=models.Annotation,
        )

    def delete(self, annotation_id: str) -> models.DeleteResponse:
        """Deletes a specific annotation from the project.

        :param annotation_id: The ID of the annotation.
        :return: A msgspec struct containing
            the deleted annotation ID and the deletion status with the following structure:

            .. code-block:: python

                DeleteResponse(deleted=True, id='annot_8188782f-a86b-4961-9e2a-697509085460')

        :example:
            .. code-block:: python

                from datature.nexus import Client

                project = Client("5aa41e8ba........").get_project("proj_b705a........")
                project.annotations.delete("asset_6aea3395-9a72-4bb5-9ee0-19248c903c56")
        """
        assert isinstance(annotation_id, str)
        return self.requester.DELETE(
            f"/projects/{self.project_id}/annotations/{annotation_id}",
            response_type=models.DeleteResponse,
        )

    def create_export(
        self,
        export_metadata: Union[AnnotationExportMetadata, dict],
        background: bool = False,
    ) -> models.Operation:
        """Exports all annotations from the project in a specific annotation format.

        :param export_metadata: A dict containing other export options.
        :param background: Signal to complete the annotation export
                            process in the background. Defaults to False.
        :return: A msgspec struct containing the operation metadata of the annotation export
            with the following structure:

            .. code-block:: python

                Operation(
                    id='op_ea10c06e-0e2a-4087-b2cf-1a1c9c42d83d',
                    kind='nexus.annotations.export',
                    status=OperationStatus(
                        overview='Finished',
                        message='Operation finished',
                        progress=OperationProgress(
                            unit='whole operation',
                            with_status=OperationProgressWithStatus(
                                Queued=0,
                                Running=0,
                                Finished=1,
                                Cancelled=0,
                                Errored=0
                            )
                        )
                    ),
                    create_date=1701927649302,
                    update_date=1701927649302
                )

        :example:
            .. code-block:: python

                from datature.nexus import Client

                project = Client("5aa41e8ba........").get_project("proj_b705a........")

                project.annotations.create_export({
                    "format": "csv_fourcorner",
                    "options": {
                        "split_ratio": 0.5,
                        "seed": 1337,
                        "normalized": True
                    }
                })
        """
        assert isinstance(export_metadata, (AnnotationExportMetadata, dict))
        assert isinstance(background, bool)

        if isinstance(export_metadata, dict):
            export_metadata = AnnotationExportMetadata(**export_metadata)

        num_annotated_assets = self.requester.GET(
            f"/projects/{self.project_id}", response_type=models.Project
        ).statistic.annotated_assets

        if num_annotated_assets == 0:
            raise error.BadRequestError(
                "There are no annotations in the project. Please annotate some assets first."
            )
        if num_annotated_assets == 1:
            export_metadata.options.split_ratio = 0.0

        operation = self.requester.POST(
            f"/projects/{self.project_id}/annotationexports",
            request_body=export_metadata.to_json(),
            response_type=models.Operation,
        )

        if background:
            return operation

        return self.operation.wait_until_done(operation.id)

    def download_exported_file(
        self, op_id: str, path: Union[str, Path, None] = None
    ) -> models.LocalAnnotations:
        """Retrieves the download link of the exported annotations.

        :param op_id: The operation ID of the annotation export.
        :param path: The download path for the annotation, default current path.
        :return: A msgspec struct with the download metadata of
            the annotation export with the following structure:

            .. code-block:: python

                LocalAnnotations(
                    download_path='local',
                    file_names=[
                        'cd067221d5a6e4007ccbb4afb5966535/train.csv',
                        'cd067221d5a6e4007ccbb4afb5966535/validate.csv'
                    ]
                )

        :example:
            .. code-block:: python

                from datature.nexus import Client

                project = Client("5aa41e8ba........").get_project("proj_b705a........")

                project.annotations.download_exported_file(
                    "op_cf8c538a-bcb5-49a9-82cf-fb0d13b49bb1"
                )
        """
        assert isinstance(op_id, str)
        assert isinstance(path, (str, Path, type(None)))

        exported_files = self.requester.GET(
            f"/projects/{self.project_id}/annotationexports/{op_id}",
            response_type=models.ExportedAnnotations,
        )

        download_path = utils.get_download_path(path)

        if exported_files.status == "Finished" and exported_files.download is not None:
            download_tmpfile = utils.download_files_to_tempfile(exported_files.download)

            with zipfile.ZipFile(download_tmpfile, "r") as zip_ref:
                zip_ref.extractall(download_path)

                file_names = [
                    info.filename for info in zip_ref.infolist() if not info.is_dir()
                ]

            return msgspec.json.decode(
                json.dumps(
                    {"download_path": str(download_path), "file_names": file_names}
                ),
                type=models.LocalAnnotations,
            )

        # If didn't find the model key in the artifacts, raise an 404 error
        raise error.NotFoundError(
            f"Export with id {op_id} not found, please export first."
        )

    def create_import_session(self, background: bool = False):
        """
        Creates an import session to import or update annotations.
            For bulk annotation import, we allow the user to add up to 1000
            files in one single import session.
            To add file into the import session, simply include its file path as an argument
            when calling the add function. Once all files have been added,
            it will initiate the upload process automatically.

        :param background: A flag indicating whether the upload should run in the background.
            Default is False.
        :return: ImportSession class
        :example:

            .. code-block:: python

                from datature.nexus import Client

                project = Client("5aa41e8ba........").get_project("proj_b705a........")

                # Import session context manager
                # If background is set to True,
                # the import session will be processed in the background

                import_session = project.annotations.create_import_session(background=True)

                with import_session as session:
                    # Function 1: add file path
                    session.add_path("path/to/annotation1.json")

                    # Function 2: add folder path
                    session.add_path("folder/path/to/files")

                    # Function 3: or even you can add bytes data
                    session.add_bytes(b"bytes/of/file", "filename")

                # Retrieve import session logs
                import_session.get_logs()

        """
        return ImportSession(self._context, background)

    def list_import_sessions(self) -> models.ImportSessions:
        """Lists the import sessions from the project.

        :return: A list of msgspec struct containing import sessions with the following structure:

            .. code-block:: python

                [ImportSession(
                    id='annotsess_58bb96a6-9dd5-4c78-89e7-57366060318e',
                    project_id='proj_cd067221d5a6e4007ccbb4afb5966535',
                    status=ImportSessionStatus(
                        overview='Finished',
                        message='Annotation Imported successfully.',
                        update_date=1701677312962,
                        annotations=ImportSessionStatusAnnotations(
                            with_status=ImportSessionAnnotationStatus(
                                Processed=1548,
                                Committed=1548
                            )
                        ),
                        files=ImportSessionStatusFiles(
                            total_size_bytes=63368,
                            page_count=1,
                            with_status=ImportSessionFilesStatus(
                                Processing=0,
                                Processed=1,
                                FailedProcess=0
                            )
                        )
                    ),
                    expiry_date=1701927649302,
                    create_date=1701927649302,
                    update_date=1701927649302
                )]

        :example:

            .. code-block:: python

                from datature.nexus import Client

                project = Client("5aa41e8ba........").get_project("proj_b705a........")

                project.annotations.list_import_sessions()
        """
        return self.requester.GET(
            f"/projects/{self.project_id}/annotationimportsessions",
            response_type=models.ImportSessions,
        )

    def get_import_session(self, import_session_id: str) -> models.ImportSession:
        """Retrieves a specific import session from the project.

        :param import_session_id: The ID of the import session.

        :return: A msgspec struct containing the import session with the following structure:

            .. code-block:: python

                ImportSession(
                    id='annotsess_58bb96a6-9dd5-4c78-89e7-57366060318e',
                    project_id='proj_cd067221d5a6e4007ccbb4afb5966535',
                    status=ImportSessionStatus(
                        overview='Finished',
                        message='Annotation Imported successfully.',
                        update_date=1701677312962,
                        annotations=ImportSessionStatusAnnotations(
                            with_status=ImportSessionAnnotationStatus(
                                Processed=1548,
                                Committed=1548
                            )
                        ),
                        files=ImportSessionStatusFiles(
                            total_size_bytes=63368,
                            page_count=1,
                            with_status=ImportSessionFilesStatus(
                                Processing=0,
                                Processed=1,
                                FailedProcess=0
                            )
                        )
                    ),
                    expiry_date=1701927649302,
                    create_date=1701927649302,
                    update_date=1701927649302
                )

        :example:

            .. code-block:: python

                from datature.nexus import Client

                project = Client("5aa41e8ba........").get_project("proj_b705a........")

                project.annotations.get_import_session(
                    "annotsess_75879caf-184d-4d82-912b-180609a72ace"
                )
        """
        assert isinstance(import_session_id, str)

        return self.requester.GET(
            f"/projects/{self.project_id}/annotationimportsessions/{import_session_id}",
            response_type=models.ImportSession,
        )

    def get_import_session_logs(
        self, import_session_id: str
    ) -> models.ImportSessionLog:
        """Retrieves import session logs from the project.

        :param import_session_id: The ID of the import session.

        :return: A msgspec struct containing the import session logs with the following structure:

            .. code-block:: python

                ImportSessionLog(
                    id='annotsess_1e433826-ab9c-402e-8b9a-a95d82f358f1',
                    project_id='proj_cd067221d5a6e4007ccbb4afb5966535',
                    logs=[]
                )

        :example:

            .. code-block:: python

                from datature.nexus import Client

                project = Client("5aa41e8ba........").get_project("proj_b705a........")

                project.annotations.get_import_session_logs(
                    "annotsess_75879caf-184d-4d82-912b-180609a72ace"
                )
        """
        assert isinstance(import_session_id, str)
        return self.requester.GET(
            f"/projects/{self.project_id}/annotationimportsessions/{import_session_id}/logs",
            response_type=models.ImportSessionLog,
        )



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   test_requester.py
@Author  :   Raighne.Weng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Datature HTTP Resource Test Cases
"""

import unittest
from unittest.mock import MagicMock

from datature.nexus import models
from datature.nexus.error import (
    BadRequestError,
    ForbiddenError,
    InternalServerError,
    NotFoundError,
    TooManyRequestsError,
    UnauthorizedError,
)
from datature.nexus.requester import Requester
from tests.unit.fixture.data import error_fixture, operation_fixture
from tests.unit.fixture.mock import MockResponse

# pylint: disable=W0703,W0012,W0212


class TestRequester(unittest.TestCase):
    """Datature HTTP Resource Test Cases."""

    def test_request_with_no_project_key(self):
        """Test resource request."""

        try:
            Requester(secret_key=None, endpoint="")._request(
                "GET",
                "test_end_point",
                response_type=models.Artifacts,
                request_body={"test": "test"},
            )

        except Exception as exception:
            assert isinstance(exception, UnauthorizedError)

    def test_request_with_request_body(self):
        """Test resource request."""
        requester = Requester("secret-key", endpoint="")
        requester._session.request = MagicMock()

        requester._session.request.side_effect = [
            MockResponse(operation_fixture.pending_operation_response, 200)
        ]
        requester._interpret_response = MagicMock()

        requester._request(
            "POST",
            "test_end_point",
            response_type=models.Operation,
            request_body={"test": "test"},
        )

    def test_request_with_query(self):
        """Test resource request."""
        requester = Requester("secret-key", endpoint="")

        requester._session.request = MagicMock()
        requester._session.request.side_effect = [
            MockResponse(operation_fixture.pending_operation_response, 200)
        ]

        response = requester.GET(
            "test_end_poimt", query={"limit": 5}, response_type=models.Operation
        )

        assert response.id == "op_c26ea31d-599c-4a03-8b0d-b2bc9995fa8a"

    def test_make_headers_with_get_method(self):
        """Test make headers."""

        headers = Requester("project_secret", endpoint="")._make_headers("GET", {})

        assert headers["Secret-Key"] == "project_secret"

    def test_make_headers_with_post_method(self):
        """Test make headers with post method."""

        headers = Requester("project_secret", endpoint="")._make_headers("POST", {})

        assert headers["Secret-Key"] == "project_secret"
        assert headers["Content-Type"] == "application/json"

    def test_make_headers_with_supplied_headers(self):
        """Test make headers with supplied headers."""

        headers = Requester("project_secret", endpoint="")._make_headers(
            "GET", {"Connection": "keep-alive", "Accept": "*/*"}
        )

        assert headers["Secret-Key"] == "project_secret"
        assert headers["Accept"] == "*/*"

    def test_interpret_response_with_400(self):
        """Test interpret response with 400 code."""

        try:
            Requester("secret-key", endpoint="")._interpret_response(
                MockResponse(error_fixture.forbidden_error_response, 400)
            )

        except Exception as exception:
            assert isinstance(exception, BadRequestError)

    def test_interpret_response_with_403(self):
        """Test interpret response with 403 code."""

        try:
            Requester("secret-key", endpoint="")._interpret_response(
                MockResponse(error_fixture.forbidden_error_response, 403)
            )

        except Exception as exception:
            assert isinstance(exception, ForbiddenError)

    def test_interpret_response_with_404(self):
        """Test interpret response with 404 code."""

        try:
            Requester("secret-key", endpoint="")._interpret_response(
                MockResponse(error_fixture.not_found_error_response, 404)
            )

        except Exception as exception:
            assert isinstance(exception, NotFoundError)

    def test_interpret_response_with_429(self):
        """Test interpret response with 429 code."""

        try:
            Requester("secret-key", endpoint="")._interpret_response(
                MockResponse(error_fixture.too_many_requests_error_response, 429)
            )

        except Exception as exception:
            assert isinstance(exception, TooManyRequestsError)

    def test_interpret_response_with_500(self):
        """Test interpret response with 500 code."""

        try:
            Requester("secret-key", endpoint="")._interpret_response(
                MockResponse(error_fixture.internal_server_error_response, 500)
            )

        except Exception as exception:
            assert isinstance(exception, InternalServerError)

    def test_make_query_string(self):
        """Test make querystring."""

        none_res = Requester("secret-key", endpoint="")._make_query_string(None)

        assert none_res == ""

        res = Requester("secret-key", endpoint="")._make_query_string(
            {
                "limit": 5,
                "offset": 0,
                "group": ["group1", "group2"],
                "includeExports": True,
            }
        )

        assert (
            res == "limit=5&offset=0&group[]=group1&group[]=group2&includeExports=true&"
        )



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   functions.py
@Author  :   Raighne.Weng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   CLI Functions
"""
# pylint: disable=E1102,R0912,C0103,R0914

import os
import re
import sys
import time
from os.path import basename, exists, isdir, join
from pathlib import Path

import requests
from alive_progress import alive_bar
from halo import Halo
from InquirerPy import inquirer

from datature import nexus
from datature.nexus import config, error
from datature.nexus.cli import messages
from datature.nexus.cli.config import Config
from datature.nexus.utils import utils


def start_spinner(fn: callable):
    """Start spinner for the function."""

    def wrapper(*args, **kwargs):
        wait_spinner = Halo(text=messages.REQUEST_SERVER_MESSAGE, spinner="dots")
        wait_spinner.start()

        try:
            result = fn(*args, **kwargs)
            return result
        finally:
            wait_spinner.stop()

    return wrapper


def get_default_datature_client():
    """
    Get default datature client from configuration file.

    :return: The default datature client.
    """
    cli_config = Config()
    project = cli_config.get_default_project()

    if project is None:
        print(messages.NO_PROJECT_MESSAGE)
        sys.exit(1)

    return nexus.Client(project.get("project_secret")).get_project(
        project.get("project_id")
    )


@start_spinner
def sdk_list_project(secret_key: str):
    """Retrieve project from server."""
    sdk_client = nexus.Client(secret_key)

    projects = sdk_client.list_projects()
    # Support to use workspace after one secret support multiple projects
    return projects


@start_spinner
def sdk_retrieve_project(project_secret: str, project_id: str):
    """Retrieve project from server."""
    sdk_project = nexus.Client(project_secret)

    project = sdk_project.get_project(project_id)

    return project


@start_spinner
def sdk_create_upload_session():
    """Retrieve project from server."""
    sdk_project = get_default_datature_client()

    import_session = sdk_project.annotations.create_import_session()

    return import_session


def sdk_retrieve_operation(op_id: str):
    """Retrieve operation from server."""
    sdk_project = get_default_datature_client()

    operation = sdk_project.operations.get(op_id)
    return operation


def download_file_from_link(link: str, download_path: str):
    """
    Download file from link.

    :param link: The url link.
    :param download_path: The path to download file.
    :return: None
    """
    query_string_removed = link.split("?")[0]
    file_name = basename(query_string_removed)

    resp = requests.get(link, stream=True, timeout=120)

    total = int(resp.headers.get("content-length", 0))
    current_size = 0
    with open(join(download_path, file_name), "wb") as file, alive_bar(
        total, title="Downloading", title_length=12, manual=True
    ) as progress_bar:
        for data in resp.iter_content(chunk_size=1024):
            size = file.write(data)
            current_size += size
            progress_bar(float(current_size / total))
        progress_bar(1.0)


def authenticate():
    """
    Authenticate the Workspace Secret with the server and creates a
    configuration file for it.

    :param secret_key: Secret key to use for the client login.
    :return: None
    """
    secret_key = inquirer.secret(message="Enter the secret key").execute().strip()

    secret_key = secret_key.strip()
    if secret_key == "":
        print(messages.AUTHENTICATION_REMINDER_MESSAGE)
        sys.exit(1)

    try:
        projects = sdk_list_project(secret_key)

        project_name = inquirer.select(
            "Which project do you want to select?",
            choices=[project.get_info().name for project in projects],
            border=True,
        ).execute()

        for project in projects:
            if project.get_info().name == project_name:
                project_id = project.get_info().id
                break
        else:
            print(messages.INVALID_PROJECT_MESSAGE)
            sys.exit(1)

        default_project = inquirer.confirm(
            f"Make [{project_name}] the default project?", default=True
        ).execute()

        cli_config = Config()
        cli_config.set_project(project_id, project_name, secret_key, default_project)

    except error.ForbiddenError:
        print(messages.INVALID_PROJECT_SECRET_MESSAGE)
        sys.exit(1)
    print(messages.AUTHENTICATION_MESSAGE)


def select_project():
    """
    Select project from saved configuration file.

    :return: None
    """
    cli_config = Config()
    project_names = cli_config.get_all_project_names()

    project_name = inquirer.select(
        "Which project do you want to select?", choices=project_names, border=True
    ).execute()
    project = cli_config.get_project_by_name(project_name)

    cli_config.set_default_project(project.get("project_id"))
    print(f"{messages.ACTIVE_PROJECT_MESSAGE}: [{project_name}]")


def list_projects():
    """
    List projects from saved configuration file.

    :return: None
    """
    cli_config = Config()
    project_names = cli_config.get_all_project_names()
    default_project = cli_config.get_default_project()

    output = [["NAME", "TOTAL_ASSETS", "ANNOTATED_ASSETS", "ANNOTATIONS", "TAGS"]]
    for project_name in project_names:
        project = cli_config.get_project_by_name(project_name)
        try:
            project = sdk_retrieve_project(
                project.get("project_secret"), project.get("project_id")
            )
            project_info = project.get_info()

            total_assets = project_info.statistic.total_assets
            annotated_assets = project_info.statistic.annotated_assets
            total_annotations = project_info.statistic.total_annotations
            tags = project_info.tags

            output.append(
                [
                    project_name,
                    total_assets,
                    annotated_assets,
                    total_annotations,
                    len(tags),
                ]
            )
        except error.ForbiddenError:
            print(
                (
                    f"Project {[project_name]} "
                    "authentication failed, please use 'datature projects auth' "
                    "again.\n"
                )
            )

    if len(output) == 0:
        print(messages.NO_PROJECT_MESSAGE)
        sys.exit(1)

    print_table(output, 20)

    active_project = default_project.get("project_name")

    print(f"\n{messages.ACTIVE_PROJECT_MESSAGE}: [{active_project}]")


def cli_loop_operation(op_id: str, data_size: int):
    """
    Upload annotations from path.

    :param path: The annotation path to upload.
    :param annotation_format: The annotation format to upload.
    :return: None
    """
    # Custom manage loop
    with alive_bar(
        data_size, title="Processing", title_length=12, length=30, manual=True
    ) as progress_bar:
        while True:
            operation = sdk_retrieve_operation(op_id)

            if operation.status.overview == "Errored":
                print(messages.UNKNOWN_ERROR_MESSAGE)
                sys.exit(1)

            count = operation.status.progress.with_status

            queued = count.Queued
            running = count.Running
            finished = count.Finished
            cancelled = count.Cancelled
            errored = count.Errored
            percentage = float(
                (int(errored) + int(cancelled) + int(finished))
                / (
                    int(errored)
                    + int(cancelled)
                    + int(finished)
                    + int(queued)
                    + int(running)
                )
            )

            if percentage == 1:
                break

            progress_bar(percentage)
            time.sleep(config.OPERATION_LOOPING_DELAY_SECONDS)
        progress_bar(1.0)
    print(messages.SERVER_COMPLETED_MESSAGE)


def upload_assets():
    """
    Upload assets from path.

    :param path: The folder to upload assets.
    :return: None
    """

    path = (
        Path(
            inquirer.filepath(
                message=messages.ASSETS_FOLDER_MESSAGE, default=os.getcwd()
            ).execute()
        )
        .expanduser()
        .resolve()
    )

    if not exists(path):
        print(messages.PATH_NOT_EXISTS_MESSAGE)
        sys.exit(1)

    # find all images under folder and sub folders
    file_paths = utils.find_all_assets(path)

    if len(file_paths) == 0:
        print(messages.NO_ASSETS_FOUND_MESSAGE)
        sys.exit(1)

    nifti_orientation = None
    if ".nii" in ",".join(str(file_path).lower() for file_path in file_paths):
        nifti_orientation = (
            inquirer.text(
                messages.ASSETS_NIFTI_DIRECTION_CHOICE_MESSAGE,
                validate=lambda x: x in ["", "x", "y", "z"],
            )
            .execute()
            .strip()
        )

    groups_res = inquirer.text(
        messages.ASSETS_GROUPS_MESSAGE,
        default="main",
        validate=lambda x: re.match(r"^[A-Za-z0-9,\s?]*$", x),
    ).execute()
    groups = [group.strip() for group in groups_res.split(",") if group.strip()]

    confirm = inquirer.confirm(
        f"""{len(file_paths)} assets will be uploaded to group(s) ({
            ', '.join(groups)})?""",
        default=True,
    ).execute()
    if not confirm:
        sys.exit(0)

    batch_size = config.ASSET_UPLOAD_SESSION_BATCH_SIZE

    num_batches = len(file_paths) // batch_size
    batches = [
        file_paths[i * batch_size : (i + 1) * batch_size] for i in range(num_batches)
    ]

    if len(file_paths) % batch_size != 0:
        batches.append(file_paths[num_batches * batch_size :])

    sdk_project = get_default_datature_client()
    # Loop Prepare asset metadata
    for _, batch in enumerate(batches):
        upload_session = sdk_project.assets.create_upload_session(
            groups=groups, background=True
        )

        with alive_bar(
            len(batch),
            length=30,
            title="Preparing",
            title_length=12,
        ) as progress_bar, upload_session as session:
            for file_path in batch:
                session.add_path(file_path, nifti_orientation=nifti_orientation)
                progress_bar()

        operations = upload_session.get_operation_ids()

        cli_loop_operation(operations[0], len(upload_session))


def upload_annotations():
    """
    Upload annotations from path.

    :param path: The annotation path to upload.
    :return: None
    """
    path = (
        Path(
            inquirer.filepath(
                message=messages.ANNOTATION_FOLDER_MESSAGE, default=os.getcwd()
            ).execute()
        )
        .expanduser()
        .resolve()
    )

    # find all images under folder and sub folders
    if isdir(path):
        file_paths = utils.find_all_annotations_files(path)
    else:
        file_paths = [path]

    confirm = inquirer.confirm(
        f"{len(file_paths)} file(s) will be uploaded?", default=True
    ).execute()
    if not confirm:
        sys.exit(0)

    batch_size = config.ANNOTATION_IMPORT_SESSION_BATCH_SIZE

    num_batches = len(file_paths) // batch_size
    batches = [
        file_paths[i * batch_size : (i + 1) * batch_size] for i in range(num_batches)
    ]

    if len(file_paths) % batch_size != 0:
        batches.append(file_paths[num_batches * batch_size :])

    # Loop Prepare asset metadata
    for _, batch in enumerate(batches):
        import_session = sdk_create_upload_session()

        with alive_bar(
            len(batch),
            length=30,
            title="Preparing",
            title_length=12,
        ) as progress_bar, import_session as session:
            for file_path in batch:
                progress_bar()

                session.add_path(file_path)

    print(messages.SERVER_COMPLETED_MESSAGE)


def download_artifact():
    """
    Download artifact model.

    :param artifact_id: The id of the artifact.
    :param model_format: The artifact model to download.
    :param path: The path to download the model.
    :return: None
    """
    path = (
        Path(
            inquirer.filepath(
                message=messages.ARTIFACT_MODEL_FOLDER_MESSAGE, default=os.getcwd()
            ).execute()
        )
        .expanduser()
        .resolve()
    )

    if not exists(path):
        print(messages.PATH_NOT_EXISTS_MESSAGE)
        sys.exit(1)

    sdk_project = get_default_datature_client()

    # call server to list all artifacts
    artifacts = sdk_project.artifacts.list(include_exports=True)
    if len(artifacts) == 0:
        print(messages.NO_ARTIFACTS_MESSAGE)
        sys.exit(1)

    artifact_lists = []
    artifacts_key_map = {}
    for artifact in artifacts:
        key = f"""{artifact.get('run_id')[-6:].upper()
                    }-{artifact.get('flow_title')}"""
        artifact_lists.append(key)
        artifacts_key_map[key] = artifact

    artifact_key = inquirer.select(
        message=messages.ARTIFACT_DOWNLOAD_MESSAGE,
        choices=artifact_lists,
        border=True,
    ).execute()
    artifact = artifacts_key_map.get(artifact_key)

    model_format = inquirer.select(
        message=messages.ARTIFACT_MODEL_FORMAT_DOWNLOAD_MESSAGE,
        choices=[option.format for option in artifact.export_options],
        border=True,
    ).execute()

    exported_formats = [export.format for export in artifact.exports]
    if model_format in exported_formats:
        # already exported, can download directly
        for model in artifact.exports:
            if model.format == model_format and model.status == "Finished":
                download_file_from_link(model.download.url, path)

    else:
        # not exported, need looping query download status
        # Loop to query status,
        wait_spinner = Halo(
            text=messages.EXPORT_ARTIFACT_WAITING_MESSAGE, spinner="dots"
        )
        wait_spinner.start()
        models = sdk_project.artifacts.create_export(
            artifact.id, {"format": model_format}
        )

        while True:
            models = sdk_project.artifacts.list_exported_models(artifact.get("id"))
            for model in models:
                if (
                    model.get("format") == model_format
                    and model.get("status") == "Finished"
                ):
                    # start download
                    wait_spinner.stop()
                    download_file_from_link(model.download.url, path)

                    return

            time.sleep(config.OPERATION_LOOPING_DELAY_SECONDS)


def download_annotations():
    """
    Export annotations from path.

    :param path: The annotation path to export.
    :param annotation_format: The annotation format to export.
    :return: None
    """
    sdk_project = get_default_datature_client()
    project_info = sdk_project.get_info()
    num_annotated_assets = project_info.statistic.annotated_assets

    if num_annotated_assets == 0:
        print(messages.NO_ANNOTATIONS_MESSAGE)
        sys.exit(1)

    path = (
        Path(
            inquirer.filepath(
                message=messages.EXPORT_ANNOTATION_FOLDER_MESSAGE,
                default=os.getcwd(),
            ).execute()
        )
        .expanduser()
        .resolve()
    )

    if not exists(path):
        print(messages.PATH_NOT_EXISTS_MESSAGE)
        sys.exit(1)

    annotations_formats = utils.get_exportable_annotations_formats(project_info.type)
    annotation_format = inquirer.select(
        message=messages.ANNOTATION_FORMAT_MESSAGE,
        choices=annotations_formats,
        border=True,
    ).execute()

    normalized = inquirer.confirm(
        message=messages.DOWNLOAD_ANNOTATIONS_NORMALIZED_MESSAGE
    ).execute()

    if num_annotated_assets > 1:
        split_ratio = round(
            int(
                inquirer.number(
                    message=messages.DOWNLOAD_ANNOTATIONS_SPLIT_RATIO_MESSAGE,
                    default=0,
                    min_allowed=0,
                    max_allowed=100,
                ).execute()
            )
            / 100,
            2,
        )
    else:
        split_ratio = 0.0

    # Loop to query status
    wait_spinner = Halo(text=messages.ANNOTATION_DOWNLOAD_MESSAGE, spinner="dots")
    wait_spinner.start()
    operation = sdk_project.annotations.create_export(
        {
            "format": annotation_format,
            "options": {
                "split_ratio": split_ratio,
                "seed": 1337,
                "normalized": normalized,
            },
        }
    )

    wait_spinner.stop()

    cli_loop_operation(operation.id, 1)

    sdk_project.annotations.download_exported_file(operation.id, path)

    print(messages.ANNOTATION_DOWNLOADED_MESSAGE)


def print_table(data: [[str]], column_width: int = 16):
    """
    List assets group statistics.

    :param data: The element array.
    :param column_width: The column widths and separator characters
    :return: None
    """
    # Print the table header
    print("".join(f"{item:{column_width}}" for _, item in enumerate(data[0])))

    # Print the table data
    for row in data[1:]:
        print("".join(f"{str(item):{column_width}}" for _, item in enumerate(row)))


def assets_group():
    """
    List assets group statistics.

    :param group: The name of group.
    :return: None
    """
    sdk_project = get_default_datature_client()

    project = sdk_project.get_info()

    groups = project.groups
    if groups is None or len(groups) == 0:
        print(messages.NO_ASSETS_GROUP_MESSAGE)
        sys.exit(1)

    selected_groups = inquirer.select(
        messages.CHOOSE_GROUP_MESSAGE,
        choices=groups,
        multiselect=True,
        default=groups,
        border=True,
    ).execute()

    statistics = sdk_project.assets.list_groups(selected_groups)

    table = [["NAME", "TOTAL", "ANNOTATED", "REVIEW", "TOFIX", "COMPLETED"]]
    for result in statistics:
        table.append(
            [
                result.group,
                result.statistic.total_assets,
                result.statistic.annotated_assets,
                result.statistic.reviewed_assets,
                result.statistic.to_fixed_assets,
                result.statistic.completed_assets,
            ]
        )
    print_table(table)



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   common.py
@Author  :   Wei Loon Cheng, Kai Xuan Lee
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Custom runner common functions.
"""

import shlex
import subprocess
import sys
import traceback
from datetime import datetime

from datature.nexus.cli.runner import commands as COMMANDS
from datature.nexus.cli.runner import consts as CONSTS
from datature.nexus.cli.runner import messages as MESSAGES


def install_pip_package(package_name: str):
    """Install pip package.

    :param package_name (str): Package name.
    :return (bool): True if successful, False otherwise.ise.
    """
    result = subprocess.run(
        shlex.split(
            COMMANDS.PIP_INSTALL_PACKAGE_COMMAND.format(sys.executable, package_name)
        ),
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        check=True,
        text=True,
    )
    if result.returncode != 0:
        return False
    return True


def log_errors(unknown: bool = False) -> None:
    """Log errors to file.

    :param unknown (bool): True if error is unknown, False otherwise.
    """
    CONSTS.ERROR_LOG_DIR.mkdir(parents=True, exist_ok=True)
    now = datetime.now()
    logs_path = (
        CONSTS.ERROR_LOG_DIR / f"datature-{now.strftime('%Y-%m-%d_%H-%M-%S')}.log"
    )
    logs_path.write_text(traceback.format_exc())
    if unknown:
        print(MESSAGES.RUNNER_UNKNOWN_ERROR_MESSAGE.format(logs_path))
    else:
        print(MESSAGES.RUNNER_TRACEBACK_MESSAGE.format(logs_path))



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   config.py
@Author  :   Raighne.Weng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   CLI config class
"""

import base64
import io
import os
import sys
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import yaml

from .messages import INVALID_PROJECT_MESSAGE, NO_PROJECT_MESSAGE


class Config:
    """Handles YAML configuration files"""

    def __init__(self):
        """Init config file"""

        self._path = self._get_config_path()
        self._data = self._parse()

    def _get_config_path(self) -> Path:
        """Get config file path"""
        if sys.platform == "win32":
            config_path = Path.home() / "AppData/Roaming/Datature/config.yaml"
        else:
            config_path = Path.home() / ".config/datature/config.yaml"

        if os.getenv("DATATURE_SDK_CONFIG"):
            config_path = Path(os.getenv("DATATURE_SDK_CONFIG"))

        config_path.parent.mkdir(parents=True, exist_ok=True)

        return config_path

    def _parse(self) -> Dict[str, Any]:
        """Parses config from YAML configuration file"""

        if not self._path:
            return {}
        try:
            with open(self._path, "r", encoding="utf8") as stream:
                return yaml.safe_load(stream)
        except FileNotFoundError:
            return {}

    def _get(self, key: str, default: Optional[Any] = None) -> Dict[str, Any]:
        """Get JSON value from self._data"""

        data: Any = self._data.copy()

        while True:
            if isinstance(key, str):
                key = key.split("/")
            key, *keys = key

            data = data.get(key)

            if data is None:
                return default
            if len(keys) == 0:
                return data
            key = keys

    def _save(self) -> None:
        """Persist the config to the file system"""

        if not self._path:
            return
        with io.open(self._path, "w", encoding="utf8") as config_file:
            yaml.dump(
                self._data, config_file, default_flow_style=False, allow_unicode=True
            )

    def set(self, key: Union[str, List[str]], value: Any, save: bool = True) -> None:
        """
        Sets the value for the specified key.

        :param key: The key where the value is going to be stored.
        :param value: The value to be stored.
        :param save: If ``True``, persists the value in the FileSystem. Defaults to ``True``.
        :return: None
        """
        if isinstance(key, str):
            key = key.split("/")

        pointer = self._data

        for k in key[:-1]:
            pointer = pointer.setdefault(k, {})
        pointer[key[-1]] = str(value)

        if save:
            self._save()

    def set_project(
        self,
        project_id: str,
        project_name: str,
        project_secret: str,
        default_project: bool,
    ) -> None:
        """
        Stores the project secret and project information.

        :param project_id: The id of the project.
        :param project_name: The name of the project.
        :param project_secret: The secret of the project.
        :param default_project: If set this project as the default project.
        :return: None
        """
        self.set(
            f"project/{project_id}/project_secret",
            base64.b64encode(project_secret.encode("ascii")).decode("ascii"),
        )
        self.set(f"project/{project_id}/project_name", project_name)

        if default_project or len(self._get("project").keys()) == 1:
            self.set("default_project", project_id)

    def get_all_project_names(self) -> List[str]:
        """
        list all saved projects name

        :return: [str]
        """
        projects = self._get("project")
        if projects is None:
            print(NO_PROJECT_MESSAGE)
            sys.exit(1)

        project_ids = projects.keys()
        projects = []

        for project_id in project_ids:
            projects.append(self._get(f"project/{project_id}/project_name"))

        return projects

    def get_default_project(self) -> dict:
        """
        list all saved projects

        :return: [str]
        """
        project_id = self._get("default_project")

        if project_id is None:
            print(NO_PROJECT_MESSAGE)
            sys.exit(1)

        return {
            "project_name": self._get(f"project/{project_id}/project_name"),
            "project_id": project_id,
            "project_secret": base64.b64decode(
                self._get(f"project/{project_id}/project_secret").encode("ascii")
            ).decode("ascii"),
        }

    def set_default_project(self, project_id: str) -> None:
        """
        Set default project

        :param project_id: The id of the project.
        :return: None
        """
        self.set("default_project", project_id)

    def get_project(self, project_id: str) -> dict:
        """
        get selected project

        :param project_id: The id of the project.
        :return: dict
        """
        return self._get(f"project/{project_id}")

    def get_project_by_name(self, project_name: str) -> dict:
        """
        get project by project name

        :param project_name: The name of the project.
        :return: dict
        """
        project_ids = list(self._get("project").keys())

        for project_id in project_ids:
            if project_name == self._get(f"project/{project_id}/project_name"):
                return {
                    "project_name": self._get(f"project/{project_id}/project_name"),
                    "project_id": project_id,
                    "project_secret": base64.b64decode(
                        self._get(f"project/{project_id}/project_secret").encode(
                            "ascii"
                        )
                    ).decode("ascii"),
                }

        print(INVALID_PROJECT_MESSAGE)
        sys.exit(1)



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   dicom_processor.py
@Author  :   Raighne.Weng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Dicom Processor
"""

import tempfile
from os import makedirs, path
from pathlib import Path
from typing import List

import cv2
from pydicom import config, dcmread

from datature.nexus import error

from .base_processor import BaseProcessor

# Enforcing Valid DICOM
config.settings.reading_validation_mode = config.RAISE


class DicomProcessor(BaseProcessor):
    """DICOM processor class"""

    def valid(self, request_data):
        """Valid the input

        :param request_data: The request data, include file path.
        :return: None.
        """
        file_path = request_data.get("file")

        if not file_path:
            raise error.BadRequestError("Required field")

    def process(self, request_data) -> List[str]:
        """Start process file to asset video

        :param request_data: The request data, include file path.
        :return: str: The generate video path.
        """
        out_path = tempfile.mkdtemp()
        file_path = request_data.get("file")

        file_name = Path(file_path).stem

        video_output = path.join(out_path, file_name)

        if not path.exists(video_output):
            makedirs(video_output)

        dicom_data = dcmread(file_path)

        four_cc = cv2.VideoWriter_fourcc(*"mp4v")

        number_of_frames = int(dicom_data.get("NumberOfFrames", 1))

        # if only contain one image, convert to image.
        if number_of_frames == 1:
            cv2.imwrite(f"{video_output}/{file_name}.png", dicom_data.pixel_array)

            return [f"{video_output}/{file_name}.png"]

        video_writer = cv2.VideoWriter(
            f"{video_output}/{file_name}.mp4",
            four_cc,
            30.0,
            (dicom_data.Rows, dicom_data.Columns),
        )

        for _, image in enumerate(dicom_data.pixel_array):
            new_image = cv2.cvtColor(image, cv2.COLOR_GRAY2RGB)
            video_writer.write(new_image)

        video_writer.release()

        return [f"{video_output}/{file_name}.mp4"]



#!/usr/bin/env python
# -*-coding:utf-8 -*-
"""
  ████
██    ██   Datature
  ██  ██   Powering Breakthrough AI
    ██

@File    :   test_annotation_import_session.py
@Author  :   Raighne.Weng
@Contact :   developers@datature.io
@License :   Apache License 2.0
@Desc    :   Datature Asset Upload Session API Test Cases
"""
# pylint: disable=W0703,R0913,W0613,W0212

import json
import unittest
from unittest.mock import ANY, MagicMock, patch

import msgspec

from datature.nexus import models
from datature.nexus.api.annotation.import_session import ImportSession
from datature.nexus.client_context import ClientContext
from datature.nexus.error import Error
from tests.unit.fixture.data import operation_fixture


class TestAnnotationImportSession(unittest.TestCase):
    """Datature Annotation Import Session API Resource Test Cases."""

    def __init__(self, *args, **kwargs):
        """init global variables."""
        unittest.TestCase.__init__(self, *args, **kwargs)

        context = ClientContext("secret_key", endpoint="http://localhost:8080")
        context.project_id = "project_id"
        context.requester._request = MagicMock()
        self.context = context

    def test_add_with_file_not_exist(self):
        """Test not existed annotation to import session."""
        import_session = ImportSession(self.context)

        try:
            import_session.add_path("assetPath")
        except Exception as exception:
            assert isinstance(exception, FileNotFoundError)

    def test_get_logs(self):
        """Test get_logs ."""
        import_session = ImportSession(self.context)

        import_session.import_session_id = "import_session_id"

        import_session.get_logs()

        import_session._context.requester._request.assert_called_with(
            "GET",
            "/projects/project_id/annotationimportsessions/import_session_id/logs",
            ANY,
            query=None,
            request_params=None,
        )

    def test_patch_status(self):
        """Test patch_status function."""
        import_session = ImportSession(self.context)

        import_session.import_session_id = "import_session_id"

        import_session._patch_status(
            {
                "condition": "FilesProcessed",
                "status": "FailedReach",
            }
        )

        import_session._context.requester._request.assert_called_with(
            "PATCH",
            "/projects/project_id/annotationimportsessions/import_session_id",
            ANY,
            query=None,
            request_params=None,
            request_body={
                "status": {
                    "conditions": [
                        {
                            "condition": "FilesProcessed",
                            "status": "FailedReach",
                        }
                    ]
                }
            },
        )

    def test_import_session_length(self):
        """Test len(import_session) function."""
        import_session = ImportSession(self.context)

        import_session.filenames = {"file1", "file2"}

        assert len(import_session) == 2

    @patch("datature.nexus.api.annotation.import_session.open")
    @patch("datature.nexus.api.annotation.import_session.path")
    @patch("datature.nexus.api.annotation.import_session.utils")
    def test_add_path_to_import_session(
        self, patched_utils, patched_path, patched_open
    ):
        """Test import session."""
        import_session = ImportSession(self.context)

        patched_open.__enter__.return_value.__len__.return_value = 100
        patched_path.isdir.return_value = True
        patched_path.basename.side_effect = ["file1", "file2"]
        patched_utils.find_all_annotations_files.return_value = ["file1", "file2"]

        import_session._calculate_file_hash = MagicMock()
        import_session._calculate_file_hash.return_value = 1234

        import_session.add_path("assetPath")

    @patch("datature.nexus.api.annotation.import_session.open")
    @patch("datature.nexus.api.annotation.import_session.path")
    @patch("datature.nexus.api.annotation.import_session.utils")
    def test_duplicate_path_to_import_session(
        self, patched_utils, patched_path, patched_open
    ):
        """Test import session."""
        import_session = ImportSession(self.context)

        patched_open.__enter__.return_value.__len__.return_value = 100
        patched_path.isdir.return_value = True
        patched_path.basename.side_effect = ["file1", "file1"]
        patched_utils.find_all_annotations_files.return_value = ["file1", "file2"]

        import_session._calculate_file_hash = MagicMock()
        import_session._calculate_file_hash.return_value = 1234

        try:
            import_session.add_path("assetPath")
        except Exception as exception:
            assert isinstance(exception, Error)

    @patch("datature.nexus.api.annotation.import_session.open")
    @patch("datature.nexus.api.annotation.import_session.path")
    @patch("datature.nexus.api.annotation.import_session.utils")
    def test_max_file_path_to_import_session(
        self, patched_utils, patched_path, patched_open
    ):
        """Test import session."""
        import_session = ImportSession(self.context)
        import_session.filenames = {"filename1", "filename2"}
        import_session.max_files_per_session = 1

        patched_path.isdir.return_value = True
        patched_utils.find_all_annotations_files.return_value = ["file1", "file2"]

        try:
            import_session.add_path("assetPath")
        except Exception as exception:
            assert isinstance(exception, Error)

    @patch("datature.nexus.api.annotation.import_session.open")
    @patch("datature.nexus.api.annotation.import_session.path")
    @patch("datature.nexus.api.annotation.import_session.utils")
    def test_background_path_to_import_session(
        self, patched_utils, patched_path, patched_open
    ):
        """Test import session."""
        import_session = ImportSession(self.context)
        import_session._upload_current_batch = MagicMock()

        patched_open.__enter__.return_value.__len__.return_value = 100
        import_session.max_files_per_batch = 1
        patched_path.isdir.return_value = True
        patched_path.basename.side_effect = ["file1", "file2"]
        patched_utils.find_all_annotations_files.return_value = ["file1", "file2"]

        import_session._calculate_file_hash = MagicMock()
        import_session._calculate_file_hash.return_value = 1234

        import_session.add_path("assetPath")
        import_session._upload_current_batch.assert_called()

    def test_add_bytes_max_files_per_session(self):
        """Test import session add bytes."""
        import_session = ImportSession(self.context)
        import_session._upload_current_batch = MagicMock()
        import_session.filenames = {"filename1"}

        import_session._calculate_file_hash = MagicMock()
        import_session._calculate_file_hash.return_value = 1234

        import_session.max_files_per_session = 1
        try:
            import_session.add_bytes(b"assetPath", "file1")
        except Exception as exception:
            assert isinstance(exception, Error)

    def test_add_bytes_duplicate_file(self):
        """Test import session add duplicate file bytes."""
        import_session = ImportSession(self.context)
        import_session._upload_current_batch = MagicMock()
        import_session.filenames = {"filename1"}

        import_session._calculate_file_hash = MagicMock()
        import_session._calculate_file_hash.return_value = 1234

        try:
            import_session.add_bytes(b"assetPath", "filename1")
        except Exception as exception:
            assert isinstance(exception, Error)

    def test_add_bytes(self):
        """Test import session add file bytes."""
        import_session = ImportSession(self.context)
        import_session._upload_current_batch = MagicMock()

        import_session._calculate_file_hash = MagicMock()
        import_session._calculate_file_hash.return_value = 1234

        import_session.add_bytes(b"assetPath", "filename1")

    def test_add_bytes_with_upload_current_batch(self):
        """Test import session add file bytes."""
        import_session = ImportSession(self.context)
        import_session._upload_current_batch = MagicMock()
        import_session.max_files_per_batch = 1
        import_session.max_bytes_per_batch = 1

        import_session._calculate_file_hash = MagicMock()
        import_session._calculate_file_hash.return_value = 1234

        import_session.add_bytes(b"assetPath", "filename1")
        import_session._upload_current_batch.assert_called()

    def test_calculate_file_hash(self):
        """Test calculate file hash."""
        import_session = ImportSession(self.context)

        c2c32 = import_session._calculate_file_hash(b"assetPath")

        assert c2c32 == "EUvjmg=="

    def test_wait_until_done(self):
        """Test wait until done."""
        import_session = ImportSession(self.context)
        import_session.import_session_id = "import_session_id"

        import_session._context.requester._request.return_value = msgspec.json.decode(
            json.dumps(operation_fixture.annotation_import_finished_response),
            type=models.ImportSession,
        )

        import_session.wait_until_done()

    @patch("datature.nexus.api.annotation.import_session.config")
    def test_wait_until_done_with_timeout(self, patched_config):
        """Test wait until done."""
        import_session = ImportSession(self.context)
        import_session.import_session_id = "import_session_id"

        patched_config.OPERATION_LOOPING_DELAY_SECONDS = 0
        patched_config.OPERATION_LOOPING_TIMEOUT_SECONDS = 2

        import_session._context.requester._request.return_value = msgspec.json.decode(
            json.dumps(operation_fixture.annotation_import_running_response),
            type=models.ImportSession,
        )

        res = import_session.wait_until_done()
        assert res is False

    def test_wait_until_done_with_error(self):
        """Test wait until done."""
        import_session = ImportSession(self.context)
        import_session.import_session_id = "import_session_id"

        import_session._context.requester._request.return_value = msgspec.json.decode(
            json.dumps(operation_fixture.annotation_import_errored_response),
            type=models.ImportSession,
        )
        try:
            import_session.wait_until_done()
        except Exception as exception:
            assert isinstance(exception, Error)
